import math
import warnings
from collections import Counter

import numpy as np
from skimage.morphology import skeletonize_3d

from .helpers import slice_normalize, min_dtype, max_dtype
from .chunky import Sparse

# optional imports

try:
    import networkx as nx

    _have_nx = True
except ImportError:
    _have_nx = False

try:
    import SimpleITK as sitk

    _have_sitk = True
except ImportError:
    _have_sitk = False

try:
    import vtk

    _have_vtk = True
except ImportError:
    _have_vtk = False


itk = None
_have_itk = False
_have_itk_thickness = False


def _try_import_itk_thickness(import_error=True):
    """Enables to import itk and itk-thickness3d on demand as it takes a long time (ca. 15s)"""
    global itk
    global _have_itk
    global _have_itk_thickness

    if _have_itk and _have_itk_thickness:
        return

    try:
        import itk

        _have_itk = True
        _have_itk_thickness = hasattr(itk, "BinaryThinningImageFilter3D")
    except ImportError:
        _have_itk = False
        _have_itk_thickness = False
    
    if not import_error:
        return

    if not _have_itk:
        raise ImportError("Please install ITK to use this function.")

    if not _have_itk_thickness:
        raise ImportError("Please install itk-thickness3d to use this function.")


def _unique_pairs(a):
    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    return np.unique(b).view(a.dtype).reshape(-1, a.shape[1])


def unique(sparse, return_counts=False, multiprocesses=1):
    """ see: https://docs.scipy.org/doc/numpy/reference/generated/numpy.unique.html """

    def process(data, prev):
        if return_counts:
            values, counts = np.unique(data, return_counts=True)
            x = dict(zip(values, counts))

            return None, dict(Counter(x) + Counter(prev))
        else:
            return None, prev | set(np.unique(data))

    result = sparse.run(
        process, prev={} if return_counts else set(), multiprocesses=multiprocesses
    )

    result_merged = {} if return_counts else set()
    for process in result:
        if return_counts:
            result_merged = dict(Counter(result_merged) + Counter(process))
        else:
            result_merged |= process
    result = result_merged
    return result


def downscale(sparse, stride=(2, 2, 2)):
    ret = Sparse.empty_like(
        sparse,
        chunks=(
            sparse.chunks[0] // stride[0],
            sparse.chunks[1] // stride[1],
            sparse.chunks[2] // stride[2],
        ),
        shape=(
            sparse.shape[0] // stride[0],
            sparse.shape[1] // stride[1],
            sparse.shape[2] // stride[2],
        ),
        spacing=(
            sparse.spacing[0] * stride[0],
            sparse.spacing[1] * stride[1],
            sparse.spacing[2] * stride[2],
        ),
    )

    for k in sparse._grid.keys():
        value = (
            sparse._grid[k][0]
            if sparse._grid[k].shape == (1,)
            else sparse._grid[k][:: stride[0], :: stride[1], :: stride[2]]
        )
        ret.set_chunk(k, value)

    return ret


def sum(sparse, multiprocesses=1):
    return np.sum(
        sparse.run(
            lambda data, prev: (None, prev + np.sum(data)),
            prev=0,
            multiprocesses=multiprocesses,
        )
    )


def downsample(sparse, stride=(3, 3, 3)):
    sparse_copy = Sparse.empty_like(sparse, chunks=np.array(stride) * 16)
    # TODO: is this copy made because of issue #16?
    sparse_copy.copy_from(sparse)

    sparse_downscaled = Sparse(
        np.array(sparse.shape) // np.array(stride), dtype=np.float32
    )
    sparse_downscaled.origin = sparse.origin
    sparse_downscaled.spacing = np.array(sparse.spacing) * np.array(stride)

    def reduce(chunk, prev):
        accu = np.zeros((16,) * 3, dtype=np.float32)

        for i in range(stride[0]):
            for j in range(stride[1]):
                for k in range(stride[2]):
                    accu += chunk[i :: stride[0], j :: stride[1], k :: stride[2]]

        coord = sparse_copy._chunk_to_global_coord(chunk.idx, (0, 0, 0)) // stride

        prev[
            coord[0] : coord[0] + 16, coord[1] : coord[1] + 16, coord[2] : coord[2] + 16
        ] = accu / (stride[0] * stride[1] * stride[2])

        # TODO: should it return (None, prev) after issue #16 ?
        return chunk, prev

    return sparse_copy.run(reduce, (0, 0, 0), sparse_downscaled)


def sum_slice(sparse, z, y, x):
    def process(idx, gs, z, y, x, prev):
        chunk_size = (
            abs(abs(x.stop - x.start) // x.step)
            * abs(abs(y.stop - y.start) // y.step)
            * abs(abs(z.stop - z.start) // z.step)
        )

        if idx in sparse._grid.keys():
            if sparse._grid[idx].shape == (1,):
                prev += sparse._grid[idx][0] * chunk_size
            else:
                prev += np.sum(sparse._grid[idx][z, y, x])
        else:
            prev += sparse.fill_value * chunk_size

        return prev

    z, y, x = slice_normalize((z, y, x), sparse.shape)

    return sparse.raw_run(
        process,
        (z.start, y.start, x.start),
        (z.stop, y.stop, x.stop),
        step=(z.step, y.step, x.step),
        prev=0,
    )


def where(sparse, func):
    ret_i, ret_j, ret_k = [], [], []

    for key in sorted(sparse._grid.keys()):
        i, j, k = np.where(func(sparse.get_chunk(key)))

        offset = sparse._chunk_to_global_coord(key, (0, 0, 0))
        ret_i.append(i + offset[0])
        ret_j.append(j + offset[1])
        ret_k.append(k + offset[2])

    if len(ret_i) == 0:
        return tuple(np.empty((0,), dtype=np.intp) for _ in range(3))

    return np.hstack(ret_i), np.hstack(ret_j), np.hstack(ret_k)


def max(sparse, multiprocesses=1):
    f = np.max
    return f(
        sparse.run(
            lambda data, prev: (None, f([prev, f(data)])),
            prev=min_dtype(sparse.dtype),
            multiprocesses=multiprocesses,
        )
    )


def nanmax(sparse, multiprocesses=1):
    f = np.nanmax
    return f(
        sparse.run(
            lambda data, prev: (None, f([prev, f(data)])),
            prev=min_dtype(sparse.dtype),
            multiprocesses=multiprocesses,
        )
    )


def max_slice(sparse, z, y, x):
    z, y, x = slice_normalize((z, y, x), sparse.shape)

    return sparse.raw_run(
        lambda idx, gs, z, y, x, prev: np.max(
            [prev, np.max(sparse.get_chunk(idx)[x, y, x])]
        ),
        (z.start, y.start, x.start),
        (z.stop, y.stop, x.stop),
        step=(z.step, y.step, x.step),
        prev=-math.inf,
    )


def min(sparse, multiprocesses=1):
    f = np.min
    return f(
        sparse.run(
            lambda data, prev: (None, f([prev, f(data)])),
            prev=max_dtype(sparse.dtype),
            multiprocesses=multiprocesses,
        )
    )


def nanmin(sparse, multiprocesses=1):
    f = np.nanmin
    return f(
        sparse.run(
            lambda data, prev: (None, f([prev, f(data)])),
            prev=max_dtype(sparse.dtype),
            multiprocesses=multiprocesses,
        )
    )


def min_slice(sparse, z, y, x):
    z, y, x = slice_normalize((z, y, x), sparse.shape)

    return sparse.raw_run(
        lambda idx, gs, z, y, x, prev: np.min(
            [prev, np.min(sparse.get_chunk(idx)[x, y, x])]
        ),
        (z.start, y.start, x.start),
        (z.stop, y.stop, x.stop),
        step=(z.step, y.step, x.step),
        prev=math.inf,
    )


def mul(sparse_a, sparse_b):
    sparse_a.run_multivariate(lambda a, b: a * b[0], [sparse_b])


def add(sparse_a, sparse_b):
    sparse_a.run_multivariate(lambda a, b: a + b[0], [sparse_b])


def subtract(sparse_a, sparse_b):
    sparse_a.run_multivariate(lambda a, b: a - b[0], [sparse_b])


def add_scalar(sparse_a, val, multiprocesses=1):
    sparse_a.run(lambda data, prev: (data + val, prev), multiprocesses=multiprocesses)


def mul_scalar(sparse_a, val, multiprocesses=1):
    sparse_a.run(lambda data, prev: (data * val, prev), multiprocesses=multiprocesses)


def any(sparse: Sparse, func=None) -> bool:
    if func is None:
        func = lambda data: data

    for key in sparse._grid.keys():
        if func(sparse.get_chunk(key)).any():
            return True

    return False


def dilate(sparse, kernel_radius, foreground_value=1, multiprocesses=1):
    """see: http://homepages.inf.ed.ac.uk/rbf/HIPR2/morops.htm

    This function requires: SimpleITK"""
    if not _have_sitk:
        raise ImportError("Please install SimpleITK to use this function.")

    if isinstance(kernel_radius, int):
        kernel_radius = (kernel_radius,) * 3

    if len(kernel_radius) == 1:
        kernel_radius = tuple([x for x in kernel_radius] * 3)

    dilate_filter = sitk.BinaryDilateImageFilter()
    dilate_filter.SetKernelRadius(kernel_radius)
    dilate_filter.SetForegroundValue(foreground_value)
    sparse.run(
        lambda data, prev: (
            sitk.GetArrayFromImage(dilate_filter.Execute(sitk.GetImageFromArray(data))),
            prev,
        ),
        envelope=kernel_radius,
        multiprocesses=multiprocesses,
    )


def erode(sparse, kernel_radius, foreground_value=1, multiprocesses=1):
    """see: http://homepages.inf.ed.ac.uk/rbf/HIPR2/morops.htm

    This function requires: SimpleITK"""
    if not _have_sitk:
        raise ImportError("Please install SimpleITK to use this function.")

    erode_filter = sitk.BinaryErodeImageFilter()
    erode_filter.SetKernelRadius(kernel_radius)
    erode_filter.SetForegroundValue(foreground_value)
    sparse.run(
        lambda data, prev: (
            sitk.GetArrayFromImage(erode_filter.Execute(sitk.GetImageFromArray(data))),
            prev,
        ),
        envelope=kernel_radius,
        skip_neighbours=True,
        multiprocesses=multiprocesses,
    )


def closing(sparse, kernel_radius, foreground_value=1, multiprocesses=1):
    """see: http://homepages.inf.ed.ac.uk/rbf/HIPR2/morops.htm

    This function requires: SimpleITK"""
    if not _have_sitk:
        raise ImportError("Please install SimpleITK to use this function.")

    closing_filter = sitk.BinaryMorphologicalClosingImageFilter()
    closing_filter.SetKernelRadius(kernel_radius)
    closing_filter.SetForegroundValue(foreground_value)
    sparse.run(
        lambda data, prev: (
            sitk.GetArrayFromImage(
                closing_filter.Execute(sitk.GetImageFromArray(data))
            ),
            prev,
        ),
        envelope=kernel_radius,
        multiprocesses=multiprocesses,
    )


def opening(sparse, kernel_radius, foreground_value=1, multiprocesses=1):
    """see: http://homepages.inf.ed.ac.uk/rbf/HIPR2/morops.htm

    This function requires: SimpleITK"""
    if not _have_sitk:
        raise ImportError("Please install SimpleITK to use this function.")

    opening_filter = sitk.BinaryMorphologicalOpeningImageFilter()
    opening_filter.SetKernelRadius(kernel_radius)
    opening_filter.SetForegroundValue(foreground_value)
    sparse.run(
        lambda data, prev: (
            sitk.GetArrayFromImage(
                opening_filter.Execute(sitk.GetImageFromArray(data))
            ),
            prev,
        ),
        envelope=kernel_radius,
        skip_neighbours=True,
        multiprocesses=multiprocesses,
    )


def thinning(sparse, envelope, multiprocesses=1):
    """1 pixel-thin wire skeletonization"""
    sparse.run(
        lambda data, prev: (
            (skeletonize_3d(data) > 0).astype(data.dtype),
            prev,
        ),
        envelope=envelope,
        skip_neighbours=True,
        multiprocesses=multiprocesses,
    )


def thinning_diameter(sparse, envelope, multiprocesses=2):
    """Computes twice the shortest distance to the outer shell along the medial axis of the object
    (i.e. diameter of the local maximal fitting sphere)

    This function requires: ITK"""
    _try_import_itk_thickness()

    if multiprocesses == 1:
        warnings.warn(
            "Disabled multiprocessing creates a pool of threads which does not get deallocated. "
            "Subsequent calls with multiprocessing **enabled** will be locked indefinitely."
        )

    sparse.run(
        lambda data, prev: (
            itk.GetArrayFromImage(
                itk.MedialThicknessImageFilter3D.New(itk.GetImageFromArray(data))
            ),
            prev,
        ),
        envelope=envelope,
        skip_neighbours=True,
        multiprocesses=multiprocesses,
    )


def to_indices_value(sparse):
    """Return non-empty (other than fill_value) elements as [(i, j, k, value), ...] array."""
    coords = []

    for k in sparse._grid.keys():
        for val in np.unique(sparse._grid[k]):
            if val != sparse.fill_value:
                z, y, x = np.where(sparse._grid[k] == val)
                offset = sparse._chunk_to_global_coord(k, (0, 0, 0))
                x += offset[2]
                y += offset[1]
                z += offset[0]
                coords.append(
                    np.dstack((z, y, x, np.full(x.shape, val))).reshape(-1, 4)
                )

    return np.vstack(coords)


def label(sparse, multiprocesses=1, fully_connected=False):
    """Label the objects in a binary image.

    ConnectedComponentImageFilter labels the objects in a binary image (non-zero pixels are considered to be objects,
    zero-valued pixels are considered to be background). Each distinct object is assigned a unique label.
    The final object labels start with 1 and are consecutive.

    :param sparse: Sparse instance
    :param multiprocesses: number of processes
    :param fully_connected:
        whether the connected components are defined strictly by face connectivity or by face+edge+vertex connectivity.
        Set `True` for objects that are 1 pixel wide.

    This function requires: NetworkX, SimpleITK
    """
    if not _have_nx:
        raise ImportError("Please install NetworkX to use this function.")
    if not _have_sitk:
        raise ImportError("Please install SimpleITK to use this function.")

    connected_compontent_filter = sitk.ConnectedComponentImageFilter()

    if multiprocesses != 1:
        raise NotImplementedError()

    if sparse.dtype != np.uint32:
        warnings.warn("label in most cases require uint32 data to work properly")

    connected_compontent_filter.SetFullyConnected(fully_connected)

    G = nx.Graph()

    def process(a, b):
        mask = np.dstack([a, b])
        relations = _unique_pairs(mask[np.sum(mask != 0, axis=2) == 2])

        for r in relations:
            G.add_edge(r[0], r[1])

    def initialize(data, prev):
        result = sitk.GetArrayFromImage(
            connected_compontent_filter.Execute(sitk.GetImageFromArray(data))
        )

        obj_count = connected_compontent_filter.GetObjectCount()
        result[result > 0] += prev[1]
        prev[0].append(obj_count)

        return result, (prev[0], prev[1] + obj_count)

    component_count_per_chunk, component_sum = sparse.run(
        initialize, prev=([], 0), multiprocesses=multiprocesses
    )[0]

    for k in sparse._grid.keys():
        # Z axis
        nk = (k[0] - 1, k[1], k[2])
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[0, :, :], sparse.get_chunk(nk)[-1, :, :])

        nk = (k[0] + 1, k[1], k[2])
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[-1, :, :], sparse.get_chunk(nk)[0, :, :])

        # Y axis
        nk = (k[0], k[1] - 1, k[2])
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[:, 0, :], sparse.get_chunk(nk)[:, -1, :])

        nk = (k[0], k[1] + 1, k[2])
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[:, -1, :], sparse.get_chunk(nk)[:, 0, :])

        # X axis
        nk = (k[0], k[1], k[2] - 1)
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[:, :, 0], sparse.get_chunk(nk)[:, :, -1])

        nk = (k[0], k[1], k[2] + 1)
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[:, :, -1], sparse.get_chunk(nk)[:, :, 0])

    cutted_G = list(G.subgraph(c) for c in nx.connected_components(G))
    val_map = {v: k + 1 for k, g in enumerate(cutted_G) for v in g}

    for idx, k in enumerate(set(range(1, component_sum + 1)) - val_map.keys()):
        val_map[k] = idx + len(cutted_G) + 1

    add_scalar(sparse, component_sum + 1)
    component_index = 1

    for k, component_count in zip(
        sorted(sparse._grid.keys()), component_count_per_chunk
    ):
        chunk = sparse.get_chunk(k)

        for _ in range(component_count):
            if component_index in val_map:
                chunk[chunk == (component_index + component_sum + 1)] = val_map[
                    component_index
                ]

            component_index += 1

        chunk[chunk > component_sum] -= component_sum + 1

        sparse.set_chunk(k, chunk)


def contour(
    sparse, values, envelope=(1, 1, 1), quantize_points_factor=0.0, multiprocesses=1
):
    """Generate isosurface(s) from a Sparse volume.

    This function requires: VTK"""
    if not _have_vtk:
        raise ImportError("Please install vtk library to use this function.")

    from .vtk_utils import add_np_to_vti
    from .geometry.extract import triangles

    def chunk_contour(data, prev, *args):
        envelope, values = args

        vti = vtk.vtkImageData()
        vti.SetOrigin(data.origin)
        vti.SetDimensions(np.array(data.shape) - np.array(envelope))
        vti.SetSpacing(data.spacing)
        series_name = "sparse"
        add_np_to_vti(
            vti, data[: -envelope[0], : -envelope[1], : -envelope[2]], series_name
        )
        vti.GetPointData().SetActiveScalars(series_name)
        contour = vtk.vtkImageMarchingCubes()
        contour.SetInputData(vti)
        contour.SetNumberOfContours(values.shape[0])
        for i, v in enumerate(values):
            contour.SetValue(i, v)
        contour.ComputeScalarsOn()
        contour.Update()

        writer = vtk.vtkPolyDataWriter()
        writer.SetInputData(contour.GetOutput())
        writer.WriteToOutputStringOn()
        writer.Update()
        prev.append(writer.GetOutputString())

        # TODO: could this return (None, prev) after issue #16 ?
        return data, prev

    out = sparse.run(
        chunk_contour,
        envelope,
        np.array([values]).flatten(),
        envelope=envelope,
        multiprocesses=multiprocesses,
        prev=[],
    )

    appendFilter = vtk.vtkAppendPolyData()
    for contour in out:
        for c in contour:
            reader = vtk.vtkPolyDataReader()
            reader.ReadFromInputStringOn()
            reader.SetInputString(c)
            reader.Update()
            appendFilter.AddInputData(reader.GetOutput())

    quantize = vtk.vtkQuantizePolyDataPoints()
    quantize.SetQFactor(quantize_points_factor)

    quantize.SetInputConnection(appendFilter.GetOutputPort())
    quantize.Update()

    return triangles(quantize.GetOutput())



import logging
import os
import warnings

import vtk
from vtk.util import numpy_support

import numpy as np

logging.debug(("VTK version:", vtk.VTK_VERSION))


# region VTK I/O

# Great explanation of VTK file and data formats are in
# "VTK Textbook" and "VTK User Guide" chapter 12: "Readers" and 19: "VTK File Formats".


def read_vtk(path, return_reader=False):
    path = os.path.abspath(path)
    logging.debug('Reading: "{}".'.format(path))

    _, file_extension = os.path.splitext(path)

    if file_extension.endswith(".vti"):
        reader = vtk.vtkXMLImageDataReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".stl"):
        reader = vtk.vtkSTLReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".vtp"):
        reader = vtk.vtkXMLPolyDataReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".vtu"):
        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".mhd"):
        reader = vtk.vtkMetaImageReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".vtk"):
        reader = vtk.vtkUnstructuredGridReader()
        reader.SetFileName(path)
        reader.ReadAllVectorsOn()
        reader.ReadAllScalarsOn()
    else:
        raise Exception("Unsupported file extension.")

    reader.Update()

    if not return_reader:
        return reader.GetOutput()
    return reader.GetOutput(), reader


def save_vti(vti, path):
    _save(vti, path, vtk.vtkXMLImageDataWriter)


def save_vtp(vtp, path):
    _save(vtp, path, vtk.vtkXMLPolyDataWriter)


def save_vtu(vtu, path):
    _save(vtu, path, vtk.vtkXMLUnstructuredGridWriter)


def save_stl(vtp, path):
    _save(vtp, path, vtk.vtkSTLWriter)


def _save(input_data, path, writer_type):
    path = os.path.abspath(path)
    logging.debug('Saving: "{}".'.format(path))
    writer = writer_type()
    writer.SetFileName(path)
    writer.SetInputData(input_data)
    writer.Write()


# endregion

# region Conversions


def add_np_to_vti(vti, arr_np, arr_name, arr_type=None):
    with warnings.catch_warnings():
        # FutureWarning:
        # Conversion of the second argument of issubdtype from `complex` to `np.complexfloating` is deprecated.
        # In future, it will be treated as `np.complex128 == np.dtype(complex).type`.
        warnings.simplefilter("ignore")
        arr = numpy_support.numpy_to_vtk(
            num_array=arr_np.ravel(), deep=True, array_type=arr_type
        )

    arr.SetName(arr_name)
    vti.GetPointData().AddArray(arr)


def vti_to_np(vti, array, components=1):
    x, y, z = vti.GetDimensions()
    arr_np = numpy_support.vtk_to_numpy(vti.GetPointData().GetArray(array))
    if components == 1:
        return arr_np.reshape(z, y, x)
    else:
        return arr_np.reshape(z, y, x, components)


def vtp_to_np(vtp, arrays):
    if vtp.GetPolys().GetMaxCellSize() > 3:
        cut_triangles = vtk.vtkTriangleFilter()
        cut_triangles.SetInputData(vtp)
        cut_triangles.Update()
        vtp = cut_triangles.GetOutput()

    result = [
        (
            "vertices",
            numpy_support.vtk_to_numpy(vtp.GetPoints().GetData()).astype(np.float32),
        ),
        (
            "triangles",
            numpy_support.vtk_to_numpy(vtp.GetPolys().GetData()).reshape(-1, 4)[:, 1:4],
        ),
    ]
    result.extend(
        [
            (arr, numpy_support.vtk_to_numpy(vtp.GetPointData().GetArray(arr)))
            for arr in arrays
        ]
    )
    return dict(result)


def vti_to_string(vti):
    writer = vtk.vtkXMLImageDataWriter()
    writer.SetInputData(vti)
    writer.WriteToOutputStringOn()  # must be before Update call
    writer.Update()
    return writer.GetOutputString()


def vtp_to_string(vtp):
    polydata_writer = vtk.vtkPolyDataWriter()
    polydata_writer.WriteToOutputStringOn()
    polydata_writer.SetInputData(vtp)
    polydata_writer.Update()
    return polydata_writer.GetOutputString()


def string_to_vtp(s):
    reader = vtk.vtkPolyDataReader()
    reader.ReadFromInputStringOn()
    reader.SetInputString(s)
    reader.Update()
    return reader.GetOutput()


def sailfish_vti_to_npy(vti_file, verbose=False, rho_name="rho", v_name="v"):
    """
    Read vti files from sailfish and return
    gets only rho and v
    """
    vti, reader = read_vtk(vti_file, return_reader=True)
    # info = reader.GetInformation()

    field_names = [
        reader.GetPointArrayName(i) for i in range(reader.GetNumberOfPointArrays())
    ]
    logging.getLogger(vti_file).debug(("fields:", field_names))
    assert rho_name in field_names and v_name in field_names

    data_rho = vti_to_np(vti, rho_name)
    data_v = vti_to_np(vti, v_name, components=3)
    data_v = np.rollaxis(data_v, 3, 0)
    # alt.  data = data.transpose(2,1,0)

    return data_rho, data_v


# endregion

# region SimpleITK interop.


def read_sitk(path):
    import SimpleITK as sitk

    path = os.path.abspath(path)
    logging.debug('Reading "{}".'.format(path))
    return sitk.ReadImage(path)


def save_sitk(img_sitk, path):
    import SimpleITK as sitk

    path = os.path.abspath(path)
    logging.debug('Saving "{}".'.format(path))
    sitk.WriteImage(img_sitk, path, False)


def save_sitk_as_vti(img_sitk, path, array_name="data"):
    vti = sitk_to_vti(img_sitk, array_name)
    save_vti(vti, path)


def sitk_to_vti(img, array_name, array_type=None):
    import SimpleITK as sitk

    vti = vtk.vtkImageData()
    vti.SetSpacing(img.GetSpacing())
    vti.SetOrigin(img.GetOrigin())
    vti.SetDimensions(img.GetSize())

    voxels = sitk.GetArrayFromImage(img)
    arr = numpy_support.numpy_to_vtk(
        num_array=voxels.ravel(), deep=True, array_type=array_type
    )
    arr.SetName(array_name)
    vti.GetPointData().SetScalars(arr)

    return vti


def vti_to_sitk(vti, array):
    import SimpleITK as sitk

    arr_np = vti_to_np(vti, array)
    arr_sitk = sitk.GetImageFromArray(arr_np)
    arr_sitk.SetOrigin(vti.GetOrigin())
    arr_sitk.SetSpacing(vti.GetSpacing())
    return arr_sitk


# endregion

# region Some strange functions


def probe_vt(vtX_file, point_data, fname=None, shape2d=None, verbose=False):
    """get values interpolated from vtu mesh on the set of points (N,3)"""
    vtX = read_vtk(vtX_file)

    points = vtk.vtkPoints()
    for point in point_data:
        points.InsertNextPoint(*point)

    polydata = vtk.vtkPolyData()
    polydata.SetPoints(points)

    probe = vtk.vtkProbeFilter()
    probe.SetSourceData(vtX)
    probe.SetInputData(polydata)
    probe.Update()

    out = probe.GetOutput()
    # out.GetBounds()
    # to get points: numpy_support.vtk_to_numpy(out.GetPoints().GetData()).shape
    pd = out.GetAttributesAsFieldData(0)
    log = logging.getLogger(vtX_file)
    if fname is None:
        # all fields
        output = dict()
        for i in range(pd.GetNumberOfArrays()):
            v_interp_on_grid = numpy_support.vtk_to_numpy(pd.GetArray(i))
            if shape2d:
                v_interp_on_grid = v_interp_on_grid.reshape(shape2d)
            log.debug(("appending in output:", pd.GetArrayName(i)))
            output[pd.GetArrayName(i)] = v_interp_on_grid
        assert len(output) > 0
        return output
    else:
        field_numbers = [
            i for i in range(pd.GetNumberOfArrays()) if pd.GetArrayName(i) == fname
        ]
        assert len(field_numbers) == 1
        log.debug(("output:", pd.GetArrayName(field_numbers[0])))
        v_interp_on_grid = numpy_support.vtk_to_numpy(pd.GetArray(field_numbers[0]))
        if shape2d:
            return v_interp_on_grid.reshape(shape2d)
        return v_interp_on_grid


def probe_vtu(
    vtu_file="output.vtu", point_data=[[0.050640, 0.027959, 0.05213]], *args, **kwargs
):
    """get values interpolated from vtu mesh on the set of points (N,3)"""
    probe_vt(vtu_file, *args, **kwargs)


def probe_vti(
    vti_file="output.vti", point_data=[[0.050640, 0.027959, 0.05213]], *args, **kwargs
):
    """get values of interpolated from vti file mesh on the set of points (N,3)"""
    probe_vt(vti_file, *args, **kwargs)


def mod_mesh(du, stlfile="c0006.stl", output_fn="surface.vtp", write=False, sign=1):
    """
    Move stl in normal direction
    """

    if type(stlfile) is str:
        stl = read_vtk(stlfile)
    else:
        stl = stlfile

    vertices = numpy_support.vtk_to_numpy(stl.GetPoints().GetData())
    # indices = numpy_support.vtk_to_numpy(stl.GetPolys().GetData()).reshape(-1, 4)[:, 1:4]

    merged = vtk.vtkPolyData()
    merged.DeepCopy(stl)

    # Compute normals to vertices
    normalGenerator = vtk.vtkPolyDataNormals()
    normalGenerator.SetInputData(merged)
    normalGenerator.ComputePointNormalsOn()
    normalGenerator.ComputeCellNormalsOff()
    normalGenerator.SetSplitting(0)
    normalGenerator.SetConsistency(0)
    normalGenerator.Update()

    merged = normalGenerator.GetOutput()
    normals = numpy_support.vtk_to_numpy(merged.GetPointData().GetNormals())

    points = vtk.vtkPoints()

    for normal, pos in zip(normals, vertices):
        points.InsertNextPoint(pos + normal * (-sign * du))

    merged.SetPoints(points)

    return merged


def xyz_at_dx(du=0.001, stlfile="c0006.stl", sign=1):
    """
    Returns equidistant points from stl in normal direction
    """

    if type(stlfile) is str:
        stl = read_vtk(stlfile)
    else:
        stl = stlfile
    # stl = stlImageActor(stlfile)
    vertices = numpy_support.vtk_to_numpy(stl.GetPoints().GetData())
    indices = numpy_support.vtk_to_numpy(stl.GetPolys().GetData()).reshape(-1, 4)[
        :, 1:4
    ]
    merged = vtk.vtkPolyData()
    merged.DeepCopy(stl)

    # Compute normals to vertices
    normalGenerator = vtk.vtkPolyDataNormals()
    normalGenerator.SetInputData(merged)
    normalGenerator.ComputePointNormalsOn()
    normalGenerator.ComputeCellNormalsOff()
    normalGenerator.SetSplitting(0)
    normalGenerator.SetConsistency(0)
    normalGenerator.Update()

    merged = normalGenerator.GetOutput()
    normals = numpy_support.vtk_to_numpy(merged.GetPointData().GetNormals())

    points = []
    for normal, pos in zip(normals, vertices):
        points.append(pos + normal * (-sign * du))

    return np.array(points)


def probe_at_dx(
    du=0.001,
    velocity_file=None,
    stlfile="c0006.stl",
    output_fn="surface.vtp",
    velocity_name="v [m/s]",
    write=False,
    mu=1.0,
    move_mesh=False,
    sign=1,
):
    """
    Equidistant points from stl in normal direction
    """

    if type(stlfile) is str:
        stl = read_vtk(stlfile)
    else:
        stl = stlfile

    vertices = numpy_support.vtk_to_numpy(stl.GetPoints().GetData())
    # indices = numpy_support.vtk_to_numpy(stl.GetPolys().GetData()).reshape(-1, 4)[:, 1:4]
    merged = vtk.vtkPolyData()
    merged.DeepCopy(stl)

    vel_data = read_vtk(velocity_file)

    # Compute normals to vertices
    normalGenerator = vtk.vtkPolyDataNormals()
    normalGenerator.SetInputData(merged)
    normalGenerator.ComputePointNormalsOn()
    normalGenerator.ComputeCellNormalsOff()
    normalGenerator.SetSplitting(0)
    normalGenerator.SetConsistency(0)
    normalGenerator.Update()

    merged = normalGenerator.GetOutput()
    normals = numpy_support.vtk_to_numpy(merged.GetPointData().GetNormals())

    points = vtk.vtkPoints()
    pointsPolyData = vtk.vtkPolyData()

    for normal, pos in zip(normals, vertices):
        points.InsertNextPoint(pos + normal * (-sign * du))

    pointsPolyData.SetPoints(points)
    probe_filter = vtk.vtkProbeFilter()
    probe_filter.SetInputData(pointsPolyData)
    probe_filter.SetSourceData(vel_data)
    probe_filter.GetOutputPort()
    probe_filter.Update()

    probed_data = probe_filter.GetOutput().GetPointData()

    if isinstance(velocity_name, str):
        v_vec = numpy_support.vtk_to_numpy(probed_data.GetArray(velocity_name))
    elif isinstance(velocity_name, list):
        vx = numpy_support.vtk_to_numpy(probed_data.GetArray(velocity_name[0]))
        vy = numpy_support.vtk_to_numpy(probed_data.GetArray(velocity_name[1]))
        vz = numpy_support.vtk_to_numpy(probed_data.GetArray(velocity_name[2]))
        v_vec = np.stack((vx, vy, vz), 1).reshape((-1, 3))
    else:
        raise NotImplementedError

    logging.debug(v_vec.shape)

    velocity = vtk.vtkFloatArray()
    velocity.SetNumberOfComponents(1)
    velocity.SetName("X_at_epsilon")

    for v in v_vec:
        # if np.max(v) > 1e33 or np.min(v) < -1e33:
        #     s = np.array([0.00])

        velocity.InsertNextTypedTuple([v])

    merged.GetPointData().AddArray(velocity)

    if move_mesh:
        merged.SetPoints(points)
        logging.debug("moving point by dx inside")

    merged.Modified()
    if write:
        save_vtp(merged, output_fn)

    return merged


def scale_and_trans(
    vtk_data=None,
    output=None,
    scale=1000.0,
    deltaxyz=[-14.52326308, 180.637182, 161.81502267],
):
    """
    Performs scaling (e.g from meters to mm) and translation of the dataset.
    Note that `vtk_data` is reader.GetOutput()
    """
    transform = vtk.vtkTransform()
    transform.Scale(scale, scale, scale)
    transform.Translate(*deltaxyz)
    transformFilter = vtk.vtkTransformFilter()
    transformFilter.SetTransform(transform)
    transformFilter.SetInputData(vtk_data)
    transformFilter.Update()
    if output is None:
        return transformFilter.GetOutput()
    else:
        save_vtu(transformFilter.GetOutput(), output)


def vtp_from_verts_faces(verts, faces):
    """
    Make vtp polydata object from vertices and indices/faces
    """

    poly = vtk.vtkPolyData()

    Points = vtk.vtkPoints()
    Points.SetData(numpy_support.numpy_to_vtk(verts.astype(np.float32)))

    vtk_id_array = numpy_support.numpy_to_vtk(
        np.pad(
            faces.astype(np.int64), [(0, 0), (1, 0)], mode="constant", constant_values=3
        ).flatten(),
        array_type=vtk.VTK_ID_TYPE,
    )

    vtk_cells = vtk.vtkCellArray()
    vtk_cells.SetCells(faces.shape[0], vtk_id_array)

    poly.SetPoints(Points)
    poly.SetPolys(vtk_cells)
    return poly


from scipy.ndimage.filters import gaussian_filter
from skimage.measure import marching_cubes


def marching_cubes_with_smooth(mask, sigma=1.2, level=0.5):
    """
    Compute isosurface around mask, converts also boundaing box.

    Paramteters
    -----------
    mask  : Sparse mask
    sigma : blur radius (in lattice units)
    level : isolevel after gaussian blurring (e.g. 0.5 for binary mask)

    Returns
    -------
    verts, faces  : mesh vertices and triangles.

    """
    from .chunky import Sparse

    origin, spacing = mask.origin, mask.spacing
    mask_s = Sparse.empty_like(mask, dtype=np.float32)
    mask_s.copy_from(mask)
    mask_s.run(lambda d, _: (gaussian_filter(d, sigma), None), envelope=(5, 5, 5))
    verts, faces, normals, values = marching_cubes(mask_s[:, :, :], 0.5)
    verts = np.array(origin, dtype=np.float32) + (
        verts[:, ::-1].astype(np.float32) * spacing
    )

    return verts, faces


# endregion



import vtk


def _extract(polydata, celltypes):
    ids = vtk.vtkIdTypeArray()
    ids.SetNumberOfComponents(1)

    for i in range(polydata.GetNumberOfCells()):
        c = polydata.GetCell(i)
        if c.GetCellType() in celltypes:
            ids.InsertNextValue(i)

    sn = vtk.vtkSelectionNode()
    sn.SetFieldType(vtk.vtkSelectionNode.CELL)
    sn.SetContentType(vtk.vtkSelectionNode.INDICES)
    sn.SetSelectionList(ids)

    sel = vtk.vtkSelection()
    sel.AddNode(sn)

    es = vtk.vtkExtractSelection()
    es.SetInputData(0, polydata)
    es.SetInputData(1, sel)
    es.Update()

    gf = vtk.vtkGeometryFilter()
    gf.SetInputConnection(es.GetOutputPort())
    return gf


def triangles(polydata):
    gf = _extract(
        polydata,
        celltypes=[
            vtk.VTK_TRIANGLE,
        ],
    )
    gf.Update()
    return gf.GetOutput()



import numpy as np


class Chunk(np.ndarray):
    def __new__(cls, input_array, origin=None, spacing=None):
        obj = np.asarray(input_array).view(cls)
        obj.origin = origin
        obj.spacing = spacing
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.origin = getattr(obj, "origin", None)
        self.spacing = getattr(obj, "spacing", None)

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, val):
        self._spacing = val

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, val):
        self._origin = val

    def to_dict(self):
        return {b"origin": self.origin, b"spacing": self.spacing, b"ndarray": self}

    @staticmethod
    def from_dict(d):
        return Chunk(d[b"ndarray"].copy(), origin=d[b"origin"], spacing=d[b"spacing"])



import unittest

import numpy as np

from chunky3d import Sparse
from chunky3d.sparse_func import (
    contour,
    dilate,
    label,
    mul_scalar,
    thinning,
    to_indices_value,
    unique,
    where,
    _have_itk,
    _have_itk_thickness,
    _have_nx,
    _have_sitk,
    _have_vtk,
)
import chunky3d.sparse_func as sf


@unittest.skipUnless(_have_vtk, "this test needs VTK")
class TestContour(unittest.TestCase):
    def test_contour(self):
        shape = (100, 100, 100)
        sp = Sparse(shape=shape)
        step = 0

        step += 1
        sp[50, 50, 50] = 100
        c = contour(sp, 100)
        self.assertEqual(c.GetNumberOfCells(), 0)
        self.assertEqual(c.GetNumberOfPoints(), 0)

        step += 1
        sp[50, 50, 50] = 101
        c = contour(sp, 100)
        self.assertEqual(c.GetNumberOfPoints(), 6)
        self.assertEqual(c.GetNumberOfCells(), 8)


class TestFunctions(unittest.TestCase):
    def test_to_indices_value(self):
        sp = Sparse(shape=(100, 100, 100))
        sp[1, 2, 3] = 4
        sp[50, 60, 70] = 80
        result = to_indices_value(sp)
        result.sort(axis=0)
        np.testing.assert_array_equal(
            result,
            [[1, 2, 3, 4], [50, 60, 70, 80]],
            verbose=True,
        )

    def test_sum(self):
        sp = Sparse(shape=(10, 10, 10))
        sp[0, 2, 1] = 2
        sp[4, 3, 5] = 4
        self.assertEqual(sf.sum(sp), 6)

    def test_max(self):
        sp = Sparse(shape=(10, 10, 10))
        sp[0, 2, 1] = 2
        sp[4, 3, 5] = 4
        self.assertEqual(sf.max(sp), 4)
        sp[4, 4, 3] = np.nan
        self.assertTrue(np.isnan(sf.max(sp)))
        self.assertEqual(sf.nanmax(sp), 4)

    def test_min(self):
        sp = Sparse(shape=(10, 10, 10))
        sp[0, 2, 1] = 2
        sp[4, 3, 5] = 4
        self.assertEqual(sf.min(sp), 0)
        sp[0, 2, 2] = -2
        sp[4, 3, 3] = -4
        self.assertEqual(sf.min(sp), -4)
        sp[4, 4, 3] = np.nan
        self.assertTrue(np.isnan(sf.min(sp)))
        self.assertEqual(sf.nanmin(sp), -4)

    def test_unique(self):
        sp = Sparse(shape=(10, 10, 10))
        sp[0, 2, 1] = 2
        sp[4, 3, 5] = 4
        self.assertSetEqual(unique(sp), {0, 2, 4})

    def test_where_with_values(self):
        sp = Sparse(shape=(10, 10, 10), chunks=4)
        sp[0, 2, 1] = 1
        sp[5, 5, 5] = 2
        sp[7, 1, 9] = 3
        result = where(sp, lambda x: x > 1)
        expected = np.array([5, 7]), np.array([1, 5]), np.array([5, 9])
        np.testing.assert_array_equal(np.sort(result, axis=1), expected)

    def test_where_empty(self):
        sp = Sparse(shape=(10, 10, 10), chunks=4)
        result = where(sp, lambda x: x > 1)
        np.testing.assert_array_equal(result, np.empty((3, 0), dtype=np.intp))

    def test_mul_scalar(self):
        sp = Sparse(shape=(4, 4, 4), chunks=2)
        sp[0, 0, 0] = 1
        sp[3, 3, 3] = 2

        mul_scalar(sp, 3)

        self.assertEqual(sp[0, 0, 0], 3)
        self.assertEqual(sp[3, 3, 3], 6)
        self.assertEqual(sp[1, 1, 1], 0)

    def test_any_no_func(self):
        sp = Sparse(shape=(4, 4, 4), chunks=2)
        sp[0, 0, 0] = 0
        
        with self.subTest("empty"):
            result_empty = sf.any(sp)
            self.assertFalse(result_empty)

        with self.subTest("not empty"):
            sp[3, 3, 3] = 2
            result_not_empty = sf.any(sp)
            self.assertTrue(result_not_empty)

    def test_any_with_func(self):
        sp = Sparse(shape=(4, 4, 4), chunks=2)
        sp[0, 0, 0] = 1
        sp[3, 3, 3] = 2
        
        with self.subTest("empty"):
            result_empty = sf.any(sp, lambda x: x > 2)
            self.assertFalse(result_empty)

        with self.subTest("not empty"):
            result_not_empty = sf.any(sp, lambda x: x > 1)
            self.assertTrue(result_not_empty)


    @unittest.skipUnless(
        _have_sitk and _have_nx, "this test needs SimpleITK and NetworkX"
    )
    def test_label(self):
        s = Sparse((30, 30, 30), dtype=np.uint32)
        s[2:5, 2:5, 2:5] = 1
        s[7:9, 7:9, 7:9] = 1
        s[15:18, 15:18, 15:18] = 1
        s[25:28, 25:28, 25:28] = 1
        label(s)
        self.assertSetEqual(unique(s), set(range(5)))

    @unittest.skipUnless(_have_sitk, "this test needs SimpleITK")
    def test_dilate(self):
        s = Sparse((15, 15, 15), dtype=np.uint8)
        s[3, 3, 3] = 1
        dilate(s, [1])  # also: dilate(s, 1) and dilate(s, (1, 1, 1))
        expected = np.array(
            [
                # as you see, corners are still empty, because this is
                # itk::BinaryBallStructuringElement (a ball kernel)
                [[0, 1, 0], [1, 1, 1], [0, 1, 0]],
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                [[0, 1, 0], [1, 1, 1], [0, 1, 0]],
            ],
            dtype=np.uint8,
        )
        np.testing.assert_array_equal(s[2:5, 2:5, 2:5], expected)
        self.assertEqual(sf.sum(s), 3 ** 3 - 8)

    def test_thinning(self):
        s = Sparse((5, 5, 5), dtype=np.uint8)
        s[...] = 1
        s[2, 2] = 0
        expected_slice = np.array(
            [
                [0, 0, 1, 0, 0],
                [0, 1, 0, 1, 0],
                [0, 1, 0, 0, 1],
                [0, 0, 1, 1, 0],
                [0, 0, 0, 0, 0],
            ],
            dtype=np.uint8,
        )
        ss = {1: s, 2: s.copy()}
        for mp in (1, 2):
            with self.subTest(multiprocesses=mp):
                ss = s.copy()
                thinning(ss, (0, 0, 0), multiprocesses=mp)
                np.testing.assert_array_equal(ss[..., 2], expected_slice)


if __name__ == "__main__":
    unittest.main()



import k3d
import numpy as np
from time import sleep


def get_k3d_obj(sparse):
    return k3d.voxels_group(
        np.array(sparse.shape, dtype=np.uint32)[::-1],
        voxels_group=sparse.get_k3d_voxels_group_dict(),
        bounds=sparse.get_voxels_bounds(),
        outlines=True,
        color_map=k3d.nice_colors[0:-2] * 50,
    )


class k3d_connector:
    k3d_voxels_chunk = None
    sparse = None

    def __init__(self, sparse):
        self.sparse = sparse
        self.k3d_voxels_chunk = dict()

        for coord in sparse.kchunks_initialized:
            self.create_k3d_chunk(coord)

    def get_ends(self, coord):
        global_coord = self.sparse._chunk_to_global_coord(coord, (0, 0, 0))
        end = np.array(self.sparse.shape) - np.array(global_coord)
        end = np.min([end, np.array(self.sparse._chunk_shape)], axis=0)

        return end

    def create_k3d_chunk(self, coord):
        end = self.get_ends(coord)
        chunk = k3d.voxel_chunk(
            voxels=self.sparse.get_chunk(coord)[: end[0], : end[1], : end[2]],
            coord=(
                np.array(coord, dtype=np.uint32)
                * np.array(self.sparse._chunk_shape, dtype=np.uint32)
            )[::-1],
            multiple=1,
        )

        self.k3d_voxels_chunk[coord] = chunk

    def get_ids(self):
        return [g.id for g in self.k3d_voxels_chunk.values()]

    def sync(self, k3d_object):
        k3d_object._hold_remeshing = True
        # to compare
        for coord in self.sparse.kchunks_initialized.intersection(
            self.k3d_voxels_chunk.keys()
        ):
            end = self.get_ends(coord)
            data = np.array(self.sparse.get_chunk(coord)[: end[0], : end[1], : end[2]])

            if np.any(self.k3d_voxels_chunk[coord].voxels != data):
                self.k3d_voxels_chunk[coord].voxels = data

        # new to add
        for coord in self.sparse.kchunks_initialized.difference(
            self.k3d_voxels_chunk.keys()
        ):
            self.create_k3d_chunk(coord)

        # new to delete
        for coord in set(self.k3d_voxels_chunk.keys()).difference(
            self.sparse.kchunks_initialized
        ):
            del self.k3d_voxels_chunk[coord]

        k3d_object.chunks_ids = self.get_ids()
        sleep(2)  # TODO: fixme race condition
        k3d_object._hold_remeshing = False



import numpy as np


def adjust_key(key, shape):
    """Prepare a slicing tuple (key) for slicing an array of specific shape."""

    if type(key) is not tuple:
        key = (key,)

    if Ellipsis in key:
        if key.count(Ellipsis) > 1:
            raise IndexError("an index can only have a single ellipsis ('...')")

        ellipsis_pos = key.index(Ellipsis)

        missing = (slice(None),) * (len(shape) - len(key) + 1)
        key = key[:ellipsis_pos] + missing + key[ellipsis_pos + 1 :]

    if len(key) > len(shape):
        raise IndexError("too many indices for array")

    if len(key) < len(shape):
        missing = (slice(None),) * (len(shape) - len(key))
        key += missing

    return key


def _check_slice(slc):
    if any(
        x is not None and not np.issubdtype(type(x), np.integer)
        for x in (slc.start, slc.stop, slc.step)
    ):
        raise IndexError("slice indices must be integers or None")


def slice_normalize(key, shape):
    """change None values and negative indices to values in range <0; shape>

    :param key: tuple of slices, argument for __setitem__ or __getitem__
    :param shape: shape of array
    :return: tuple of 'normalized' slices
    :raises: IndexError if slice start, stop or step is not an integer"""

    shape = np.asarray(shape, dtype=np.int_)
    key = adjust_key(key, shape)
    k = list()

    for ctr, slc in enumerate(key):
        if not isinstance(slc, slice):  # when key is integer
            if not np.issubdtype(type(slc), np.integer):
                raise IndexError("single indices must be integers")

            if slc < -shape[ctr] or slc >= shape[ctr]:
                raise IndexError(
                    f"index {slc} is out of bounds for axis {ctr} with size {shape[ctr]}"
                )

            if slc < 0:
                # nasty case for [-1]
                slc + shape[ctr]

            slc = slice(slc, slc + 1, 1)

        _check_slice(slc)

        start, stop, step = slc.start, slc.stop, slc.step

        if step is None:
            step = 1
        if step == 0:
            raise ValueError(f"slice step of axis {ctr} cannot be zero")

        if start is None:
            start = 0 if step > 0 else shape[ctr] - 1
        else:
            # bring start to (-shape[ctr], shape[ctr])
            start = min(max(start, -shape[ctr]), shape[ctr])
            if start < 0:
                start = shape[ctr] + start

        if stop is None:
            stop = shape[ctr] if step > 0 else -1
        else:
            # bring stop to (-shape[ctr], shape[ctr])
            stop = min(max(stop, -shape[ctr]), shape[ctr])
            if stop < 0:
                stop = shape[ctr] + stop

        k.append(slice(start, stop, step))
    return tuple(k)


def slice_shape(key, array_shape):
    """Determine effective shape of slicing an array

    :param key: tuple of slices, argument for __setitem__ or __getitem__
    :param shape: shape of array
    :return: resulting shape"""

    key = slice_normalize(key, array_shape)
    shape = list()

    for slc in key:
        if (slc.stop - slc.start) * slc.step <= 0:
            # range direction must match step sign
            # and slc.start != slc.stop to have a nonempty
            shape.append(0)
            continue

        span = abs(slc.stop - slc.start)  # regardless of which is greater
        step = abs(slc.step)  # dirction already compatible
        shape.append((span + step - 1) // step)  # __ceildiv__

    return tuple(shape)


def pad_to_chunk(a, n):
    """
    Pad dense array to multiplicity of a number (e.g. chunk size)
    """
    to_n = lambda x, n: (0, x + (n - x % n) % n - x)
    to_ = lambda x: to_n(x, n)
    a_padded = np.pad(a, list(map(to_, a.shape)), mode="constant")
    return a_padded


def check_start_end(start, end, shape, check_end=True):
    if (start < np.zeros(3)).any():
        raise IndexError(f"start index {start} out of range {shape}.")
    # if |step| is > 1, testing the end is not as simple
    if check_end and (end > np.array(shape)).any():
        raise IndexError(f"end index {end} out of range {shape}.")


def min_dtype(t):
    """Get the minimum value for the given dtype."""
    if issubclass(t, np.inexact):
        return np.finfo(t).min
    else:
        return np.iinfo(t).min


def max_dtype(t):
    """Get the maximum value for the given dtype."""
    if issubclass(t, np.inexact):
        return np.finfo(t).max
    else:
        return np.iinfo(t).max


def tighten_box(geo):
    """Remove empty space from bounding box in sparse"""

    from .chunky import Sparse

    nonempty_chunks = {
        idx: c for idx, c in geo._grid.items() if not np.all(c == geo.fill_value)
    }

    origins = np.array([c.origin for idx, c in nonempty_chunks.items()])
    idxs = np.array([idx for idx, c in nonempty_chunks.items()])

    geo_tight = Sparse(
        (1 + np.max(idxs, axis=0) - np.min(idxs, axis=0)) * np.array(geo.chunks),
        chunks=geo.chunks,
        dtype=geo.dtype,
        origin=np.min(origins, axis=0),
        spacing=geo.spacing,
    )
    for idx, c in nonempty_chunks.items():
        idx_from_zero = tuple(np.array(idx) - np.min(idxs, axis=0))
        geo_tight.set_chunk(idx_from_zero, c)

    return geo_tight



from chunky3d import Sparse, point_probe

import unittest
import os

import numpy as np


class TestImage(unittest.TestCase):
    def test_zero(self):
        shp = (10, 10, 10)

        chunk_shape = 32

        s = Sparse(shape=np.multiply(shp, chunk_shape), chunks=chunk_shape)

        zeros = np.zeros(shape=(chunk_shape, chunk_shape, chunk_shape))

        for i in range(shp[0]):
            for j in range(shp[1]):
                for k in range(shp[2]):
                    np.testing.assert_array_equal(s.get_chunk((i, j, k)), zeros)

    def test_set_get(self):
        shp = (10, 10, 10)
        chunk_shape = 32

        dims = np.multiply(shp, chunk_shape)

        s = Sparse(shape=dims, chunks=chunk_shape)

        test_data = np.zeros(shape=dims)

        test_data[: dims[0] // 3, : dims[1] // 4, 32] = 3.14

        test_data[128:160, 128:161, 128:159] = np.random.rand(32, 33, 31)

        s.set((0, 0, 0), test_data)

        for i in range(shp[0]):
            for j in range(shp[1]):
                for k in range(shp[2]):
                    np.testing.assert_array_equal(
                        s.get_chunk((i, j, k)),
                        test_data[
                            i * chunk_shape : (i + 1) * chunk_shape,
                            j * chunk_shape : (j + 1) * chunk_shape,
                            k * chunk_shape : (k + 1) * chunk_shape,
                        ],
                    )

    def test_getitem(self):
        shp = (234, 231, 128)
        mga = np.random.rand(shp[0], shp[1], shp[2])
        s = Sparse(shp, dtype=np.float_, chunks=(16, 32, 8), fill_value=0)
        s.set((0, 0, 0), mga)

        w2 = s[-138:-10:2, -223:-39:5, 120:128:1]
        w3 = mga[-138:-10:2, -223:-39:5, 120:128:1]

        np.testing.assert_array_equal(w2, w3)

    def test_getitem_simple_vs_slice(self):
        shp = (234, 231, 128)
        mga = np.random.rand(*shp)
        s = Sparse(shp, dtype=np.float_, chunks=(16, 32, 8), fill_value=0)
        s.set((0, 0, 0), mga)

        for _ in range(10):
            i, j, k = map(np.random.randint, shp)
            v_slice = s[i : i + 1, j : j + 1, k : k + 1]
            v_simple = s[i, j, k]
            self.assertEqual(v_simple, v_slice)

    def test_setitem_simple_vs_slice(self):
        shp = (23, 12, 128)
        s_point = Sparse(shp, dtype=np.float_, chunks=(16, 32, 8), fill_value=0)
        s_slice = Sparse(shp, dtype=np.float_, chunks=(16, 32, 8), fill_value=0)

        s_point[0, 1, 2] = 3
        s_slice[0:1, 1:2, 2:3] = 3

        s_point[22, 11, 127] = 4
        s_slice[22:23, 11:12, 127:128] = 4

        for _ in range(200):
            i, j, k = map(np.random.randint, shp)
            val = np.random.random()
            s_slice[i : i + 1, j : j + 1, k : k + 1] = val
            s_point[i, j, k] = val

        np.testing.assert_array_equal(s_slice[...], s_point[...])

    def save_load(self, compression_level):
        shp = (67, 87, 33)
        mga = np.random.randint(0, 100, shp, dtype=np.uint8)
        s = Sparse(shp, dtype=np.uint8, chunks=(16, 32, 8), fill_value=0)
        s.set((0, 0, 0), mga)

        s.save("save.msgpack", compression_level)
        loaded_s = Sparse.load("save.msgpack")

        for (k1, v1), (k2, v2) in zip(vars(s).items(), vars(loaded_s).items()):
            self.assertEqual(k1, k2)

            if k1 not in ["_grid", "_memory_blocks"]:
                # fields like 'shape' etc.
                self.assertEqual(v1, v2)
                continue

            if k1 == "_grid":
                for (_, d1), (_, d2) in zip(v1.items(), v2.items()):
                    np.testing.assert_array_equal(d1, d2)
                continue

            if k1 == "_memory_blocks":
                for d1, d2 in zip(v1, v2):
                    np.testing.assert_array_equal(d1, d2)
                continue

        os.remove("save.msgpack")

    def test_save_compress(self):
        self.save_load(6)

    def test_save_without_compress(self):
        self.save_load(0)

    def test_memory_usage(self):
        s = Sparse(shape=(64, 64, 64), chunks=(2, 2, 2))
        empty_memory = s.__sizeof__()

        s[0, 0, 0] = 1
        single_chunk = s.__sizeof__()

        s[1, 1, 1] = 1
        single_chunk2 = s.__sizeof__()

        self.assertEqual(single_chunk, single_chunk2)

        s[2, 2, 2] = 1
        self.assertEqual(len(s._memory_blocks), 2)
        two_chunks = s.__sizeof__()

        self.assertAlmostEqual(
            two_chunks - single_chunk, single_chunk - empty_memory, delta=200
        )

        s.make_dense_data()
        self.assertEqual(len(s._memory_blocks), 1)
        defragmented = s.__sizeof__()

        self.assertLess(defragmented, two_chunks)

        s.update_grid_mask()
        mask_size = np.zeros(s._block_shape, dtype=np.int32).__sizeof__()
        after_grid_mask = s.__sizeof__()
        self.assertEqual(after_grid_mask, defragmented + mask_size - None.__sizeof__())
    
    def test_astype(self):
        grid_shape = (4, 4, 4)
        chunk_shape = 4
        orig_type = np.uint8
        orig_fill_value = 2
        shape = np.multiply(grid_shape, chunk_shape)
        orig_sparse = Sparse(shape=shape, chunks=chunk_shape, fill_value=orig_fill_value, dtype=orig_type)
        target_type = np.float64

        new_sparse = orig_sparse.astype(target_type)

        self.assertEqual(orig_sparse.dtype, orig_type)
        self.assertEqual(new_sparse.dtype, target_type)
        self.assertEqual(new_sparse.fill_value, np.float64(orig_fill_value))
        self.assertEqual(new_sparse.chunks, orig_sparse.chunks)
        self.assertSequenceEqual(new_sparse.shape, orig_sparse.shape)
        self.assertSequenceEqual(new_sparse.origin, orig_sparse.origin)
        self.assertSequenceEqual(new_sparse.spacing, orig_sparse.spacing)

        for i in range(grid_shape[0]):
            for j in range(grid_shape[1]):
                for k in range(grid_shape[2]):
                    self.assertEqual(new_sparse.get_chunk((i, j, k)).dtype, target_type)


class TestBroadcasting(unittest.TestCase):
    def test_too_small(self):
        # base behavior
        n = np.zeros((4, 4, 4))
        with self.assertRaises(ValueError) as cm:
            n[:3, :3, :3] = np.ones((2, 2, 2))
        self.assertTrue(str(cm.exception).startswith("could not broadcast"))

        # chunky behavior
        s = Sparse(shape=(4, 4, 4))

        with self.assertRaises(ValueError) as cm:
            s[:3, :3, :3] = np.ones((2, 2, 2))
        # time will show if this message is constant
        self.assertTrue(
            str(cm.exception).startswith("operands could not be broadcast together")
        )

    def test_too_big(self):
        # base behavior
        n = np.zeros((4, 4, 4))
        with self.assertRaises(ValueError) as cm:
            n[:3, :3, :3] = np.ones((4, 4, 4))
        self.assertTrue(str(cm.exception).startswith("could not broadcast"))

        # chunky behavior
        s = Sparse(shape=(4, 4, 4))

        with self.assertRaises(ValueError) as cm:
            s[:3, :3, :3] = np.ones((4, 4, 4))
        # time will show if this message is constant
        self.assertTrue(
            str(cm.exception).startswith("operands could not be broadcast together")
        )

    def test_uniform_value(self):
        s = Sparse(shape=(4, 4, 4))
        s[...] = 3
        np.testing.assert_array_equal(s[...], np.full((4, 4, 4), 3))
        s[...] = 9
        np.testing.assert_array_equal(s[...], np.full((4, 4, 4), 9))

    def test_step_broadcast(self):
        expected = np.zeros((3, 3, 3))
        expected[:2] = [3, 4, 5]
        s = Sparse(shape=(3, 3, 3))
        s[:2] = [3, 4, 5]
        np.testing.assert_array_equal(s[...], expected)

        expected = np.zeros((3, 3, 3))
        expected[::2] = [3, 4, 5]
        s = Sparse(shape=(3, 3, 3))
        s[::2] = [3, 4, 5]
        np.testing.assert_array_equal(s[...], expected)


class TestInterpolation(unittest.TestCase):
    def setUp(self):
        self.s = Sparse(shape=(3, 3, 3), chunks=(2, 2, 2))
        self.s[1, 1, 1] = 1.0
        self.s[2, 2, 2] = 2.0
        self.s.make_dense_data()
        self.s.update_grid_mask()

    def test_grid_mask_exception(self):
        s = Sparse(shape=(2, 2, 2), chunks=(2, 2, 2))
        s[1, 1, 1] = 1.0
        # RuntimeError: Missing grid_mask in sparse array. Use update_grid_mask() before point_probe().
        with self.assertRaises(RuntimeError):
            point_probe(np.zeros((1, 3)), s)

    def test_non_contiguous_memory_blocks(self):
        s = Sparse(shape=(3, 3, 3), chunks=(2, 2, 2))
        s[1, 1, 1] = 1.0
        s[2, 2, 2] = 1.0
        # Exception: Memory blocks have wrong len: 2, use make_dense_data() to fix.
        with self.assertRaises(Exception):
            point_probe(np.zeros((1, 3)), s)

    def test_empty_zeros(self):
        self.assertEqual(point_probe(np.zeros((1, 3)), self.s), 0.0)

    def test_flat_xyz(self):
        self.assertEqual(point_probe(np.ones(3), self.s), 1.0)

    def test_middle_of_voxel(self):
        self.assertEqual(point_probe(np.ones((1, 3)) * 0.5, self.s), 0.125)

    def test_list_xyz(self):
        self.assertEqual(point_probe([0.5, 0.5, 0.5], self.s), 0.125)

    def test_unequal_spacing(self):
        s = Sparse(shape=(2, 2, 2), chunks=(2, 2, 2), spacing=(1, 2, 4))
        s[:2, :2, :2] = np.arange(8).reshape((2, 2, 2))
        s.update_grid_mask()
        self.assertEqual(point_probe([0.5, 0.5, 0.5], s), 1.5)

    def test_shape_corner_probe(self):
        self.assertEqual(point_probe(np.full(3, 2), self.s), 2.0)


class TestProperties(unittest.TestCase):
    def test_unsigned_shape(self):
        s = Sparse(tuple(np.arange(2, 5, dtype=np.uint8)))
        self.assertTrue(np.all(-np.array(s.shape) < 0))  # unsinged would be > 0

    def test_negative_shape(self):
        with self.assertRaises(ValueError) as cm:
            Sparse((10, -2, 10))
        self.assertTrue(
            str(cm.exception).startswith("all dimensions of shape must be positive")
        )


if __name__ == "__main__":
    unittest.main()



__all__ = [
    "Sparse",
    "point_probe",
    "Chunky",
]

from .chunky import (
    Sparse,
    point_probe,
)

from ._version import (
    __version__,
    version_info,
)

# compatibility aliases:
Chunky = Sparse



import platform
import zlib

from numba import jit, njit, prange
import dill
import msgpack
import msgpack_numpy as m
import numpy as np
import psutil
import scipy.ndimage

from .chunk import Chunk
from .helpers import slice_normalize, slice_shape, check_start_end
from .multiprocesses import ProcessPool


@njit
def fast_get(
    ijk, dense_data, grid_mask, chunks, envelope=0, fill_value=0.0, error_value=0.0
):
    if ijk[0] < 0 or ijk[1] < 0 or ijk[2] < 0:
        return error_value

    bi, li = divmod(ijk[0], chunks[0])
    bj, lj = divmod(ijk[1], chunks[1])
    bk, lk = divmod(ijk[2], chunks[2])

    shp = grid_mask.shape

    if bi > shp[0] or bj > shp[1] or bk > shp[2]:
        return error_value

    idx = grid_mask[bi, bj, bk]

    if idx != 0xFFFFFFFF:
        return dense_data[
            envelope + idx * (chunks[0] + 2 * envelope) + li,
            envelope + lj,
            envelope + lk,
        ]
    else:
        return fill_value


@njit
def fast_get_interpolated(ijk, dense_data, grid_mask, shape, envelope=0):
    """
         *-----------*
        /|5         /|6           k (x)
       / |         / |            ^
      /  |        /  |            |
     *-----------*   |            |
     |8  *-------|7--*             ----> i (z)
     |  / 1      |  / 2          /
     | /         | /            /
     |/          |/            v
     *-----------*            j (y)
    4           3
    """
    ijk_base = np.array(ijk).astype(np.uint32)
    t = np.array(ijk) - ijk_base

    v1 = fast_get(ijk_base, dense_data, grid_mask, shape, envelope)
    v2 = fast_get(
        ijk_base + np.array([1, 0, 0]), dense_data, grid_mask, shape, envelope
    )
    v4 = fast_get(
        ijk_base + np.array([0, 1, 0]), dense_data, grid_mask, shape, envelope
    )
    v3 = fast_get(
        ijk_base + np.array([1, 1, 0]), dense_data, grid_mask, shape, envelope
    )
    v5 = fast_get(
        ijk_base + np.array([0, 0, 1]), dense_data, grid_mask, shape, envelope
    )
    v6 = fast_get(
        ijk_base + np.array([1, 0, 1]), dense_data, grid_mask, shape, envelope
    )
    v7 = fast_get(
        ijk_base + np.array([1, 1, 1]), dense_data, grid_mask, shape, envelope
    )
    v8 = fast_get(
        ijk_base + np.array([0, 1, 1]), dense_data, grid_mask, shape, envelope
    )

    v12 = v1 * (1 - t[0]) + v2 * t[0]
    v43 = v4 * (1 - t[0]) + v3 * t[0]
    v1243 = v12 * (1 - t[1]) + v43 * t[1]

    v56 = v5 * (1 - t[0]) + v6 * t[0]
    v87 = v8 * (1 - t[0]) + v7 * t[0]
    v5687 = v56 * (1 - t[1]) + v87 * t[1]

    return v1243 * (1 - t[2]) + v5687 * t[2]


@jit(nogil=True, parallel=True)
def fast_point_probe(xyz, dense_data, grid_mask, shape, origin, spacing, envelope=0):
    """
    Interpolates Sparse on x, y, z inside bounding box.
    Expects np array xyz, of shape [Npoints,3]
    shape, origin, spacing must be np.arrays

    The proper call for p being a Sparse array is:

        fast_point_probe(xyz, p.dense_data, p.grid_mask, np.array(p.chunks),
                         np.array(p.origin), np.array(p.spacing), envelope)

           *-----------*
          /|5         /|6           x (k)
         / |         / |            ^
        /  |        /  |            |
       *-----------*   |            |
       |8  *-------|7--*             ----> z (i)
       |  / 1      |  / 2          /
       | /         | /            /
       |/          |/            v
       *-----------*            y (j)
      4           3
    """

    ijk = ((xyz - origin) / spacing)[:, ::-1]
    ijk_base_all = np.floor(ijk).astype(np.int32)
    t_all = ijk - ijk_base_all
    output = np.empty(xyz.shape[0])

    for i in prange(xyz.shape[0]):
        t = t_all[i]
        b = ijk_base_all[i]

        v1 = fast_get(b, dense_data, grid_mask, shape, envelope)
        v2 = fast_get((b[0] + 1, b[1], b[2]), dense_data, grid_mask, shape, envelope)
        v3 = fast_get(
            (b[0] + 1, b[1] + 1, b[2]), dense_data, grid_mask, shape, envelope
        )
        v4 = fast_get((b[0], b[1] + 1, b[2]), dense_data, grid_mask, shape, envelope)
        v5 = fast_get((b[0], b[1], b[2] + 1), dense_data, grid_mask, shape, envelope)
        v6 = fast_get(
            (b[0] + 1, b[1], b[2] + 1), dense_data, grid_mask, shape, envelope
        )
        v7 = fast_get(
            (b[0] + 1, b[1] + 1, b[2] + 1), dense_data, grid_mask, shape, envelope
        )
        v8 = fast_get(
            (b[0], b[1] + 1, b[2] + 1), dense_data, grid_mask, shape, envelope
        )

        v12 = v1 * (1 - t[0]) + v2 * t[0]
        v43 = v4 * (1 - t[0]) + v3 * t[0]
        v1243 = v12 * (1 - t[1]) + v43 * t[1]

        v56 = v5 * (1 - t[0]) + v6 * t[0]
        v87 = v8 * (1 - t[0]) + v7 * t[0]
        v5687 = v56 * (1 - t[1]) + v87 * t[1]

        output[i] = v1243 * (1 - t[2]) + v5687 * t[2]

    return output


def point_probe(xyz, p, envelope=0):
    """
    Wraps numba compiled `fast_point_probe`.
    """
    if p.grid_mask is None:
        raise RuntimeError(
            "Missing grid_mask in sparse array. Use update_grid_mask() before point_probe()."
        )

    xyz = np.asarray(xyz)

    if xyz.shape == (3,):
        xyz = np.expand_dims(xyz, 0)

    assert len(xyz.shape) == 2 and xyz.shape[1] == 3

    return fast_point_probe(
        xyz,
        p.dense_data,
        p.grid_mask,
        np.array(p.chunks),
        np.array(p.origin),
        np.array(p.spacing),
        envelope,
    )


class Sparse:
    """
    This is a class for storing data in chunks.
    All chunks have the same shape (each dimension of chunk must be a power of two).
    Chunks filled with unique value are stored as one value.
    Chunks filled with default value are not stored.


    Attributes:
        dtype: Data type stored in array
        shape (tuple): dimensions of data structure as dense matrix (i, j, k)
        chunks (tuple): dimensions of chunk (i, j, k)
        origin (tuple): position in world coordinates (x, y, z) of the voxel (0, 0, 0)
        spacing (tuple): distance between voxels in each direction (x, y, z)
        cdata_shape (tuple): dimensions of chunk grid (i, j, k)
        fill_value (self.dtype): value used for uninitialized (empty) portions of the array.
        nchunks (int): total number of chunks
        nchunks_initialized (int): numbers of chunks that have been initialized with some data

    Methods
        get_chunk: get chunk from grid
        set_chunk: set chunk in grid
        get: get dense submatrix from structure
        set: insert dense submatrix into structure
        get_k3d_voxels_group_dict: get chunks as list of dict for use in K3D voxels group
    """

    def __init__(
        self,
        shape,
        chunks=True,
        fill_value=0,
        dtype=np.double,
        origin=(0, 0, 0),
        spacing=(1.0, 1.0, 1.0),
    ):
        self._shape = None
        self._dtype = None
        self._chunk_shape = None
        self._block_shape = None
        self._grid_mask = None
        self._origin = None
        self._spacing = None

        self._default_value = fill_value
        self.dtype = dtype
        self.chunks = chunks
        self.shape = shape
        self.origin = origin
        self.spacing = spacing
        self._block_shape = tuple(
            np.ceil(np.divide(self.shape, self.chunks)).astype(np.int_)
        )
        self._grid = dict()

        self._memory_blocks = []
        self._memory_blocks_with_holes = 0

    def _validate_dense_data(self):
        if len(self._memory_blocks) != 1:
            raise Exception(
                "Memory blocks have wrong len: {}, use make_dense_data() to fix.".format(
                    len(self._memory_blocks)
                )
            )

        if self._memory_blocks_with_holes != 0:
            raise Exception(
                "Memory blocks having holes: {}, use make_dense_data() to fix.".format(
                    self._memory_blocks_with_holes
                )
            )

    @property
    def dense_data(self):
        self._validate_dense_data()
        return self._memory_blocks[0]

    @dense_data.setter
    def dense_data(self, val):
        self._validate_dense_data()
        self._memory_blocks[0][:] = val

    def __sizeof__(self):
        total = super().__sizeof__()

        # dicts
        total += self._grid.__sizeof__()
        for key, value in self._grid.items():
            total += key.__sizeof__()
            total += value.__sizeof__()

        # lists
        total += self._memory_blocks.__sizeof__()
        for block in self._memory_blocks:
            total += block.__sizeof__()

        # arrays
        total += self._grid_mask.__sizeof__()

        return total

    def make_dense_data(self, envelope=0):
        """Defragment self._memory_blocks into a single array with possible envelope around chunks."""

        if (
            len(self._memory_blocks) == 1
            and self._memory_blocks_with_holes == 0
            and self._memory_blocks[0].shape[2] == self._chunk_shape[2] + envelope * 2
        ):
            # already defragmented
            return

        dense_data = np.zeros(
            (
                (self._chunk_shape[0] + envelope * 2) * self.nchunks_initialized,
                (self._chunk_shape[1] + envelope * 2),
                (self._chunk_shape[2] + envelope * 2),
            ),
            dtype=self.dtype,
        )

        w = self._chunk_shape[0] + envelope * 2

        for i, k in enumerate(sorted(self._grid.keys())):
            o1, o2 = i * w + envelope, (i + 1) * w - envelope

            if envelope > 0:
                dense_data[o1:o2, envelope:-envelope, envelope:-envelope] = self._grid[
                    k
                ]
                self._grid[k] = self._make_chunk(
                    dense_data[o1:o2, envelope:-envelope, envelope:-envelope], k
                )
            else:
                dense_data[o1:o2, :, :] = self._grid[k]
                self._grid[k] = self._make_chunk(dense_data[o1:o2, :, :], k)

        del self._memory_blocks[:]
        self._memory_blocks_with_holes = 0
        self._memory_blocks = [dense_data] if dense_data.shape[0] > 0 else []

    def _global_to_chunk_coords(self, coord):
        """
        Global coordinates to chunk id and in chunk coords
        :param coord: (tuple) global coordinates
        :return: chunk_id, coord_in_chunk
        """
        return np.divmod(coord, self.chunks)

    def _chunk_to_global_coord(self, chunk_id, coord):
        """
        Local chunk coords to global coordinates
        :param chunk_id: (tuple) id of chunk in grid
        :param coord: (tuple) local coordinates in chunk
        :return: (tuple) global coordinates
        """
        ret = np.multiply(chunk_id, self.chunks)
        ret += np.array(coord)
        return ret

    @property
    def dtype(self):
        """ The NumPy data type. """
        return self._dtype

    @dtype.setter
    def dtype(self, val):
        self._dtype = val

    @property
    def shape(self):
        """A tuple of integers describing the length of each dimension of the array."""
        return self._shape

    @shape.setter
    def shape(self, val):
        if isinstance(val, int):
            val = tuple([val] * 3)
        elif isinstance(val, np.ndarray) and len(val) == 3:
            val = tuple(val.astype(np.int_))
        elif isinstance(val, tuple) and len(val) == 3:
            val = tuple(np.array(val, dtype=np.int_))
        else:
            raise ValueError("invalid shape specified")

        if any(d <= 0 for d in val):
            raise ValueError("all dimensions of shape must be positive")

        self._shape = val

    @property
    def size(self):
        return np.product(self._shape)

    @property
    def itemsize(self):
        return self._dtype.itemsize

    @property
    def chunks(self):
        """ A tuple of integers describing the length of each dimension of a chunk of the array. """
        return self._chunk_shape

    @chunks.setter
    def chunks(self, val):
        if isinstance(val, tuple) or isinstance(val, np.ndarray):
            if len(val) == 3:
                for v in val:
                    if v <= 0 and ((v & (v - 1)) != 0):
                        raise Exception("Chunk size must be a power of two")
                self._chunk_shape = tuple(val)
            else:
                raise Exception("Chunk size length error")
        elif isinstance(val, bool):
            self._chunk_shape = (32, 32, 32)
        elif isinstance(val, int):
            if val <= 0 and ((val & (val - 1)) != 0):
                raise Exception("Chunk size must be a power of two")
            self._chunk_shape = tuple([val] * 3)
        else:
            raise Exception("Chunk size dtype error")

    @property
    def cdata_shape(self):
        """ A tuple of integers describing the number of chunks along each dimension of the array. """
        return self._block_shape

    @property
    def fill_value(self):
        """ A value used for uninitialized (empty) portions of the array. """
        return self._default_value

    @property
    def nchunks(self):
        """ Total number of chunks. """
        return np.prod(self.cdata_shape)

    @property
    def nchunks_initialized(self):
        """ The number of chunks that have been initialized with some data. """
        return len(self._grid)

    @property
    def kchunks_initialized(self):
        """ List of keys of chunks that have been initialized with some data. """
        return set(self._grid.keys())

    @property
    def chunks_initialized(self):
        """ List of copies of chunks that have been initialized with some data. """
        return [
            Chunk(
                self._grid[k],
                spacing=self.spacing,
                origin=tuple(self.origin + np.multiply(k, self.spacing)),
            )
            for k in self.kchunks_initialized
        ]

    @property
    def origin(self):
        """ A tuple of floats describing the position in world coordinates (x,y,z) of the voxel (0,0,0). """
        return self._origin

    @origin.setter
    def origin(self, val):
        if isinstance(val, int) or isinstance(val, float):
            self._origin = tuple([val] * 3)
        elif isinstance(val, np.ndarray) and len(val) == 3:
            self._origin = tuple(val)
        elif isinstance(val, tuple) and len(val) == 3:
            self._origin = val
        else:
            raise Exception("Origin error!")

    @property
    def spacing(self):
        """ A tuple of floats describing the (width,height,length) of the cubical cells that compose the data set. """
        return self._spacing

    @spacing.setter
    def spacing(self, val):
        if isinstance(val, int) or isinstance(val, float):
            self._spacing = tuple([val] * 3)
        elif isinstance(val, np.ndarray) and len(val) == 3:
            self._spacing = tuple(val)
        elif isinstance(val, tuple) and len(val) == 3:
            self._spacing = val
        else:
            raise Exception("Spacing error!")

    @property
    def grid_mask(self):
        """Grid mask is a dense array indicating indices of chunks.
        It has to be manually updated using `update_grid_mask`"""
        return self._grid_mask

    def update_grid_mask(self):
        """
        An update of grid_mask attribute from dictionary of keys.
        """
        if (
            type(self._grid_mask) is not np.ndarray
            or self._grid_mask.shape != self._block_shape
        ):
            self._grid_mask = np.full(self._block_shape, 0xFFFFFFFF, dtype=np.uint32)
        else:
            self._grid_mask.fill(0xFFFFFFFF)

        self._grid_mask[tuple(np.array(sorted(self._grid.keys())).T)] = np.arange(
            self.nchunks_initialized
        )

    def get_chunk(self, idx, dtype=None):
        """
        Get chunk from the grid.
        :param:
        idx (tuple): coordinates in grid (i, j, k)
        dtype: type of the returned Chunk. Default: original dtype
        :return:
        chunk (ndarray / unique value)
        """
        if dtype is None:
            dtype = self.dtype

        # TODO: set values of chunks outside array dimensions as NaN
        if (idx < np.array([0, 0, 0])).any() or (
            idx >= np.array(self._block_shape)
        ).any():
            raise Exception("Index out of range.", idx)
        if idx in self._grid.keys():
            if self._grid[idx].shape == (1,):
                arr = np.zeros(self._chunk_shape, dtype=dtype)
                arr.fill(self._grid[idx][0])
            else:
                arr = self._grid[idx].copy()
        else:
            arr = np.zeros(self._chunk_shape, dtype=dtype)
            arr.fill(self.fill_value)

        return self._make_chunk(arr, idx)

    def _make_chunk(self, arr, idx):
        chunk = arr.view(Chunk)
        chunk.origin = (
            self.origin + (np.array(idx) * np.array(self.chunks))[::-1] * self.spacing
        )
        chunk.spacing = self.spacing
        chunk.idx = idx

        return chunk

    def _remove_memory_block(self, val):
        for i, m in enumerate(self._memory_blocks):
            if id(m) == id(val):
                del self._memory_blocks[i]

    def _add_memory_block(self, val):
        for i, m in enumerate(self._memory_blocks):
            if id(m) == id(val):
                return

        self._memory_blocks.append(val)

    def _update_memory_block(self, idx, d):
        if idx not in self._grid.keys():
            self._add_memory_block(d)
        else:
            if self._grid[idx].shape != d.shape:
                self._remove_memory_block(self._grid[idx])
                self._add_memory_block(d)

        if idx not in self._grid.keys() or self._grid[idx].shape != d.shape:
            self._grid[idx] = d
        else:
            self._grid[idx][:] = d

    def set_chunk(self, idx, val):
        """
        Set chunk in the grid.

        :param:
        idx (tuple): coordinates in grid (i, j, k)
        val (ndarray or int/float): ndarray with chunk shape (i, j, k) or unique value
        """

        # TODO: add warrning if val.dtype != self.dtype
        if (idx < np.array([0, 0, 0])).any() or (
            idx >= np.array(self._block_shape)
        ).any():
            raise IndexError(f"Index {idx} out of range {self._block_shape}.")

        if isinstance(val, np.ndarray):
            if val.shape != self._chunk_shape:
                raise ValueError("Chunk size error!")

            uniform = np.any(val != val[0, 0, 0])

            if uniform:
                d = val.astype(self.dtype)
                self._update_memory_block(idx, d)
            else:
                if np.equal(val[0, 0, 0], self.fill_value):
                    if idx in self._grid.keys():
                        if self._grid[idx].base is not None:
                            self._memory_blocks_with_holes += 1
                        else:
                            self._remove_memory_block(self._grid[idx])

                        del self._grid[idx]
                else:
                    d = val[0:1, 0:1, 0:1].astype(self.dtype).reshape((1,)).copy()
                    self._update_memory_block(idx, d)
        else:
            if np.equal(val, self.fill_value):
                if idx in self._grid.keys():
                    if self._grid[idx].base is not None:
                        self._memory_blocks_with_holes += 1
                    else:
                        self._remove_memory_block(self._grid[idx])

                    del self._grid[idx]
            else:
                d = np.array([val], dtype=self.dtype)
                self._update_memory_block(idx, d)
                self._grid[idx] = d

    def run_multivariate(self, func, sparse_list, *args):
        keys = set()

        for s in sparse_list:
            keys = keys | set(s._grid.keys())

        for k in keys:
            ret = func(self.get_chunk(k), [s.get_chunk(k) for s in sparse_list], *args)
            self.set_chunk(k, ret)

    def _run(self, *args, **kwargs):
        func = kwargs.get("sparse_func")
        envelope = kwargs.get("envelope")
        keys = kwargs.get("keys")
        prev = kwargs.get("prev")

        to_update = {}

        for k in keys:
            if envelope == (0, 0, 0):
                ret, prev = func(self.get_chunk(tuple(k)), prev, *args)
                if ret is None:
                    # read-only func
                    continue
                to_update[tuple(k)] = ret
            else:
                s = self._chunk_to_global_coord(
                    k, (-envelope[0], -envelope[1], -envelope[2])
                )
                e = self._chunk_to_global_coord(
                    k,
                    (
                        self._chunk_shape[0] + envelope[0],
                        self._chunk_shape[1] + envelope[1],
                        self._chunk_shape[2] + envelope[2],
                    ),
                )
                s_to_pad = s.copy()
                s_orig = s.copy()
                s_to_pad[s_to_pad > 0] = 0
                s_to_pad = np.abs(s_to_pad)
                s[s < 0] = 0

                e_to_pad = e.copy()
                e_to_pad[e_to_pad <= np.array(self._shape)] = 0
                e_to_pad = e_to_pad - np.array(self._shape)
                e_to_pad[e_to_pad < 0] = 0

                shp = np.array(self._shape)
                e[e > shp] = shp[e > shp]

                data = self.get(s, e)
                data = np.pad(
                    data,
                    list(zip(s_to_pad, e_to_pad)),
                    "constant",
                    constant_values=self.fill_value,
                )

                chunk = data.view(Chunk)
                chunk.spacing = self.spacing
                chunk.origin = self.origin + s_orig[::-1] * self.spacing

                ret, prev = func(chunk, prev, *args)
                if ret is None:
                    # read-only func
                    continue

                to_update[tuple(k)] = ret[
                    envelope[0] : -envelope[0],
                    envelope[1] : -envelope[1],
                    envelope[2] : -envelope[2],
                ]
        return to_update, prev

    def _get_keys_for_run(self, envelope, skip_neighbours):
        """ Helper method to determine keys which Sparse will be iterated over during `run` method """
        if envelope == (0, 0, 0):
            keys = set(self._grid.keys())
        else:
            self.update_grid_mask()
            keys = set(self._grid.keys())
            if not skip_neighbours:
                conv_size = np.array([2, 2, 2]) + np.ceil(
                    np.array(envelope) / np.array(self.chunks)
                ).astype(np.uint16)
                neighbours = scipy.ndimage.filters.convolve(
                    self.grid_mask.astype(np.uint16) != 0xFFFF,
                    np.ones(conv_size),
                    mode="constant",
                    cval=0.0,
                )
                neighbours = (
                    neighbours * ~(self.grid_mask.astype(np.uint16) != 0xFFFF)
                ) > 0
                for idx in np.dstack(np.where(neighbours == True))[0]:
                    keys.add(tuple(idx.tolist()))
        return sorted(keys)

    def _run_multiprocess(
        self,
        func,
        *args,
        envelope=(0, 0, 0),
        prev=None,
        skip_neighbours=False,
        multiprocesses=1,
    ):
        pp = ProcessPool(multiprocesses)
        to_update = {}
        keys = self._get_keys_for_run(envelope, skip_neighbours)
        for k in np.array_split(keys, multiprocesses):
            pp.add_job(
                dill.dumps(
                    {
                        "call": self._run,
                        "args": args,
                        "kwargs": {
                            "sparse_func": func,
                            "envelope": envelope,
                            "sparse": self,
                            "keys": k,
                            "prev": prev,
                        },
                    }
                )
            )
        pp.finish_pool_queue()
        prev_combined = []
        for chunks, prev in pp.shared.get():
            for key, chunk in chunks.items():
                to_update[key] = chunk
            prev_combined.append(prev)

        return to_update, prev_combined

    def _run_singleprocess(
        self, func, *args, envelope=(0, 0, 0), prev=None, skip_neighbours=False
    ):
        keys = self._get_keys_for_run(envelope, skip_neighbours)
        to_update, prev = self._run(
            *args, sparse_func=func, envelope=envelope, prev=prev, keys=keys
        )
        return to_update, [prev]  # list for compatibility with multiprocess

    def run(
        self,
        func,
        *args,
        envelope=(0, 0, 0),
        prev=None,
        skip_neighbours=False,
        multiprocesses=1,
    ):
        """
        Run given function on every initialized Chunk (pycardio.sparse.chunk).
        :param func: Function to be called on Chunk,
            always in form `func(data_in: Chunk, prev_in, *args) -> (data_out: ndarray, prev_out)` where
            `data_in` is an input Chunk,
            `prev_in` acts as an accumulator and is passed from the previous `func` execution to the next one,
            `args` is a non-keyworded variable length argument list.
            Function has to return two elements:
            `data_out` is a (possibly modified) ndarray from `data_in` which is to be inserted in-place into Sparse,
                       or `None` if nothing should be modified in the Sparse,
            `prev_out` is a (possibly modified) `prev_in` accumulator.
        :param args: non-keyworded variable length argument list which will be passed to every `func` call.
        :param envelope: 3-element tuple for extra margin on each direction in each axis on the `data_in` Chunk.
        :param prev: initial value for an accumulator.
        :param skip_neighbours: excludes chunks' neighbors from iterated keys list if `envelope` > 0.
        :param multiprocesses: number of processes spawned to execute `func` in parallel (chunk keys distributed evenly).
            For values below 0: (physical_cpus + multiprocesses + 1) is used, e.g. for -2, all **physical** CPUs but one are used.
            Multiprocessing is disabled if set to other values (specifically: 0, 1, None, ...).
        :return: a list of accumulated results.
            List's length is based on number of processes involved: [`prev_out_proc1`, `prev_out_proc2`, `prev_out_proc3`, ...]
        """

        to_update = {}
        if multiprocesses not in (0, 1):
            if platform.system() != "Linux":
                raise NotImplementedError(
                    "Multiprocessing currently works only on Unix."
                )
            if multiprocesses < 0:
                physical_cores = psutil.cpu_count(logical=False)
                multiprocesses = max(1, physical_cores + multiprocesses + 1)
            to_update, prev = self._run_multiprocess(
                func,
                *args,
                envelope=envelope,
                prev=prev,
                skip_neighbours=skip_neighbours,
                multiprocesses=multiprocesses,
            )
        else:
            to_update, prev = self._run_singleprocess(
                func,
                *args,
                envelope=envelope,
                prev=prev,
                skip_neighbours=skip_neighbours,
            )
        for k in to_update.keys():
            self.set_chunk(k, to_update[k])

        return prev

    def raw_run(self, func, start, end, prev=None, step=(1, 1, 1), *args):
        check_start_end(start, end, self._shape)

        s = np.array(start)
        e = np.array(end)
        cs = np.array(self._chunk_shape)
        blck = (
            np.floor_divide(s, self._chunk_shape),
            np.floor_divide(e, self._chunk_shape) + 1,
        )

        gs = s.copy()

        ls_c, ls, le_c, le = None, None, None, None
        for i in range(blck[0][0], blck[1][0]):
            for j in range(blck[0][1], blck[1][1]):
                for k in range(blck[0][2], blck[1][2]):
                    ls_c, ls = self._global_to_chunk_coords(gs)
                    le_c, le = self._global_to_chunk_coords(e)
                    le[ls_c != le_c] = cs[ls_c != le_c]
                    le_c[ls_c != le_c] = ls_c[ls_c != le_c]

                    if np.any(ls >= le):
                        break

                    prev = func(
                        (i, j, k),
                        gs,
                        slice(ls[0], le[0], step[0]),
                        slice(ls[1], le[1], step[1]),
                        slice(ls[2], le[2], step[2]),
                        prev,
                        *args,
                    )

                    gs[2] = (
                        self._chunk_to_global_coord(
                            ls_c, np.floor_divide(le - ls - 1, step) * step + ls
                        )[2]
                        + step[2]
                    )

                gs[2] = s[2]
                gs[1] = (
                    self._chunk_to_global_coord(
                        ls_c, np.floor_divide(le - ls - 1, step) * step + ls
                    )[1]
                    + step[1]
                )

            gs[1] = s[1]
            gs[0] = (
                self._chunk_to_global_coord(
                    ls_c, np.floor_divide(le - ls - 1, step) * step + ls
                )[0]
                + step[0]
            )

        return prev

    def get(self, start, end, step=(1, 1, 1), squeeze=True):
        """
        Get dense submatrix from structure
        :param start: (tuple) global coordinates of submatrix start (i,j,k)
        end: (tuple) global coordinates of submatrix end (i,j,k)
        :return: (ndarray) dense array
        """
        s = np.array(start)
        e = np.array(end)
        d = np.ceil(np.divide(e - s, step)).astype(np.int_)
        ret = np.empty(d, dtype=self.dtype)

        def process(idx, gs, z, y, x, _):
            tmp = self.get_chunk(idx)[z, y, x]
            ret_s = np.floor_divide(gs - s, step)

            ret[
                ret_s[0] : ret_s[0] + tmp.shape[0],
                ret_s[1] : ret_s[1] + tmp.shape[1],
                ret_s[2] : ret_s[2] + tmp.shape[2],
            ] = tmp

        self.raw_run(process, start, end, step=step)

        if squeeze:
            return np.squeeze(ret)
        else:
            return ret

    def set(self, start, val, step=(1, 1, 1)):
        """
        Insert matrix into structure

        :param:
        start (tuple): global coordinates of submatrix start (I,J,K)
        val (ndarray): matrix to be inserted in structure
        """
        if not isinstance(val, np.ndarray):
            raise Exception("Array is expected")

        s = np.array(start)
        e = s + np.array(val.shape) * step
        cs = np.array(self._chunk_shape)
        check_start_end(s, e, self._shape, check_end=(np.abs(step) == 1).all())

        blck = (
            np.floor_divide(s, self._chunk_shape),
            np.ceil(np.divide(e, self._chunk_shape)).astype(np.uint32),
        )

        gs = s.copy()
        idx = np.zeros(3, dtype=np.int_)
        ids = None
        ls_c, ls, le_c, le = None, None, None, None

        for i in range(blck[0][0], blck[1][0]):
            for j in range(blck[0][1], blck[1][1]):
                for k in range(blck[0][2], blck[1][2]):
                    ls_c, ls = self._global_to_chunk_coords(gs)
                    le_c, le = self._global_to_chunk_coords(e)
                    le[ls_c != le_c] = cs[ls_c != le_c]
                    le_c[ls_c != le_c] = ls_c[ls_c != le_c]
                    if np.any(ls >= le):
                        break

                    ids = np.floor_divide(le - ls, step)
                    tmp = val[
                        idx[0] : idx[0] + ids[0],
                        idx[1] : idx[1] + ids[1],
                        idx[2] : idx[2] + ids[2],
                    ]

                    # special case speedup - skip fill_value->fill_value
                    if (i, j, k) in self._grid.keys() or np.any(tmp != self.fill_value):
                        ret = self.get_chunk((i, j, k))
                        ret[
                            ls[0] : le[0] : step[0],
                            ls[1] : le[1] : step[1],
                            ls[2] : le[2] : step[2],
                        ] = tmp

                        self.set_chunk((i, j, k), ret)

                    idx[2] += ids[2]
                    gs[2] = (
                        self._chunk_to_global_coord(
                            ls_c, np.floor_divide(le - ls - 1, step) * step + ls
                        )[2]
                        + step[2]
                    )
                idx[2] = 0
                idx[1] += ids[1]
                gs[2] = s[2]
                gs[1] = (
                    self._chunk_to_global_coord(
                        ls_c, np.floor_divide(le - ls - 1, step) * step + ls
                    )[1]
                    + step[1]
                )
            idx[1] = 0
            idx[0] += ids[0]
            gs[1] = s[1]
            gs[0] = (
                self._chunk_to_global_coord(
                    ls_c, np.floor_divide(le - ls - 1, step) * step + ls
                )[0]
                + step[0]
            )

    def _simple_get1(self, key):
        bi, li = divmod(key[0], self._chunk_shape[0])
        bj, lj = divmod(key[1], self._chunk_shape[1])
        bk, lk = divmod(key[2], self._chunk_shape[2])

        chunk = self._grid.get((bi, bj, bk))
        if chunk is not None:
            if chunk.shape == (1,):
                return chunk[0]
            else:
                return chunk[li, lj, lk]
        else:
            if (key < np.array([0, 0, 0])).any() or (
                key >= np.array(self._shape)
            ).any():
                raise IndexError("index out of range.")
            return self.fill_value

    def __getitem__(self, key):
        if np.shape(key) == (3,) and all(isinstance(k, int) for k in key):
            return self._simple_get1(key)

        key = slice_normalize(key, self.shape)
        start = (key[0].start, key[1].start, key[2].start)
        end = (key[0].stop, key[1].stop, key[2].stop)
        step = (key[0].step, key[1].step, key[2].step)

        return self.get(start, end, step)

    def _simple_set1(self, key, val):
        bi, li = divmod(key[0], self._chunk_shape[0])
        bj, lj = divmod(key[1], self._chunk_shape[1])
        bk, lk = divmod(key[2], self._chunk_shape[2])

        chunk = self._grid.get((bi, bj, bk))
        if chunk is not None:
            if chunk.shape == (1,):
                if val != chunk[0]:
                    new_val = np.full(self._chunk_shape, chunk[0])
                    new_val[li, lj, lk] = val
                    self.set_chunk((bi, bj, bk), new_val)
                return
            else:
                chunk[li, lj, lk] = val
                return

        # not in grid
        if (key < np.array([0, 0, 0])).any() or (key >= np.array(self._shape)).any():
            raise IndexError("index out of range.")

        # new chunk
        # don't do this:
        # new_val = np.full(self._chunk_shape, self._default_value)
        new_val = self.get_chunk((bi, bj, bk))
        new_val[li, lj, lk] = val
        self.set_chunk((bi, bj, bk), new_val)

    def __setitem__(self, key, val):
        # simple case (single element)
        if (
            np.shape(val) == ()
            and np.shape(key) == (3,)
            and all(isinstance(k, int) for k in key)
        ):
            self._simple_set1(key, val)
            return

        # general case
        key = slice_normalize(key, self.shape)
        shape = slice_shape(key, self.shape)

        if np.product(shape) == 0:
            return

        if np.shape(val) != shape:
            # enable broadcasting, e.g. a single value
            val = np.broadcast_to(val, shape)

        start = tuple(k.start for k in key)
        step = tuple(k.step for k in key)
        return self.set(start, val, step)

    def crop_chunks(self, crop=((0, 0), (0, 0), (0, 0))):
        """
        Crops chunks around Sparse
        :param: number of full chunks to crop ((z_before, z_after), (y_before, y_after), (x_before, x_after))
        :returns: new Sparse object
        """
        crop = np.array(crop)
        shape = tuple(
            [
                self.chunks[dim]
                * (
                    ((self.shape[dim] - 1) // self.chunks[dim] + 1)
                    - crop[dim][0]
                    - crop[dim][1]
                )
                for dim in range(3)
            ]
        )
        origin = tuple(
            [
                self.origin[dim]
                + crop[{0: 2, 1: 1, 2: 0}[dim]][0]
                * self.chunks[dim]
                * self.spacing[dim]
                for dim in range(3)
            ]
        )
        sparse_new = Sparse.empty_like(self, shape=shape, origin=origin)

        chunk_min = crop[:, 0]
        chunk_max = sparse_new.cdata_shape + crop[:, 0]
        for k in self._grid.keys():
            if not np.any(np.array(k) < chunk_min) and not np.any(
                chunk_max < np.array(k)
            ):
                sparse_new.set_chunk(tuple(k - crop[:, 0]), self.get_chunk(k))
        return sparse_new

    def pad_chunks(self, pad=((0, 0), (0, 0), (0, 0))):
        """
        Pads chunks around Sparse
        :param: number of full chunks to pad ((z_before, z_after),(y_before, y_after), (x_before, x_after)).
                shape is *always* padded to the nearest chunk size multiple.
        :returns: new Sparse object
        """
        pad = np.array(pad)
        shape = tuple(
            [
                self.chunks[dim]
                * (
                    (self.shape[dim] - 1) // self.chunks[dim]
                    + 1
                    + pad[dim][0]
                    + pad[dim][1]
                )
                for dim in range(3)
            ]
        )
        origin = tuple(
            [
                self.origin[dim]
                - pad[{0: 2, 1: 1, 2: 0}[dim]][0] * self.chunks[dim] * self.spacing[dim]
                for dim in range(3)
            ]
        )
        sparse_new = Sparse.empty_like(self, shape=shape, origin=origin)
        for k in self._grid.keys():
            sparse_new.set_chunk(tuple(pad[:, 0] + k), self.get_chunk(k))
        return sparse_new

    def get_k3d_voxels_group_dict(self, dtype=None):
        """
        Generator returns chunks as dict for use in K3D voxels group.

        :param:
        dtype (numpy dtype): returned data dtype

        :return:
        list of dicts
        """
        if dtype is None:
            dtype = self.dtype
        ret = []
        for k in self._grid.keys():
            coord = np.array(k, dtype=np.int_) * self._chunk_shape
            mask = (coord + self._chunk_shape) > self._shape
            end = self._chunk_shape - np.multiply(
                coord + self._chunk_shape - self._shape, mask.astype(np.int_)
            )
            ret.append(
                {
                    "voxels": self.get_chunk(k)[: end[0], : end[1], : end[2]].astype(
                        dtype
                    ),
                    "coord": coord[::-1],
                    "multiple": 1,
                }
            )
        return ret

    def get_voxels_bounds(self):
        dx, dy, dz = self.spacing[0] / 2.0, self.spacing[1] / 2.0, self.spacing[2] / 2.0

        return [
            self.origin[0] - dx,
            self.origin[0] + self.spacing[0] * (self._shape[2] - 1) + dx,
            self.origin[1] - dy,
            self.origin[1] + self.spacing[1] * (self._shape[1] - 1) + dy,
            self.origin[2] - dz,
            self.origin[2] + self.spacing[2] * (self._shape[0] - 1) + dz,
        ]

    def get_bounds(self):
        return [
            self.origin[0],
            self.origin[0] + self.spacing[0] * (self._shape[2] - 1),
            self.origin[1],
            self.origin[1] + self.spacing[1] * (self._shape[1] - 1),
            self.origin[2],
            self.origin[2] + self.spacing[2] * (self._shape[0] - 1),
        ]

    def copy_from(self, sparse):
        """
        Copy **data only** from another Sparse.
        :param sparse:
        :return:
        """
        for k in sparse._grid.keys():
            coord = sparse._chunk_to_global_coord(k, (0, 0, 0))
            chunk = sparse.get_chunk(k, dtype=self.dtype)
            s = chunk.shape

            self.set(
                coord,
                chunk[
                    0 : min(s[0], self.shape[0] - coord[0]),
                    0 : min(s[1], self.shape[1] - coord[1]),
                    0 : min(s[2], self.shape[2] - coord[2]),
                ],
            )

    @staticmethod
    def empty_like(
        sparse,
        shape=None,
        chunks=None,
        fill_value=None,
        dtype=None,
        origin=None,
        spacing=None,
    ):
        return Sparse(
            shape if shape is not None else sparse.shape,
            chunks if chunks is not None else sparse.chunks,
            fill_value if fill_value is not None else sparse.fill_value,
            dtype if dtype is not None else sparse.dtype,
            origin if origin is not None else sparse.origin,
            spacing if spacing is not None else sparse.spacing,
        )

    @staticmethod
    def create_from_k3d_voxels_group_dict(k3dvg, shape=None):
        """
        Create Sparse object from K3D voxel group

        :param k3dvg: voxel group list
        :param shape: shape of the full original geometry (computed with chunk-size precision if None)
        :returns: Sparse object
        """
        geom_shape = chunk_shape = (0, 0, 0)
        for chunk in k3dvg:
            geom_shape = np.maximum(geom_shape, chunk["coord"][::-1])
            chunk_shape = np.maximum(chunk_shape, chunk["voxels"].shape)
        geom_shape += chunk_shape
        if shape is not None:
            geom_shape = shape

        s = Sparse(geom_shape, chunks=chunk_shape)

        for chunk in k3dvg:
            filler = np.zeros(chunk_shape)
            filler[
                : chunk["voxels"].shape[0],
                : chunk["voxels"].shape[1],
                : chunk["voxels"].shape[2],
            ] = chunk["voxels"]
            s.set_chunk(tuple(chunk["coord"][::-1] // chunk_shape), filler)

        return s

    def save(self, filename, compression_level=0):
        with open(filename, "wb") as file:
            d = {
                "chunks": self.chunks,
                "shape": self.shape,
                "origin": self.origin,
                "spacing": self.spacing,
                "dtype": np.dtype(self.dtype).str,
                "fill_value": self.fill_value,
                "_grid": {k: v.to_dict() for k, v in self._grid.items()},
            }

            data = msgpack.packb(d, default=m.encode)

            if compression_level > 0:
                data = zlib.compress(data, compression_level)

            file.write(data)

    @staticmethod
    def load(filename):
        with open(filename, "rb") as file:
            d = file.read()

            msgpack_args = {
                "object_hook": m.decode,
                "use_list": False,
                "strict_map_key": False,  # tuple as "map key" (msgpack 1.0)
                "raw": True,  # return keys as b'shape' (msgpack 1.0)
            }

            try:
                data = msgpack.unpackb(d, **msgpack_args)
            except msgpack.exceptions.ExtraData:
                # FIXME: not a great way to detect compression, to say the least.
                data = msgpack.unpackb(zlib.decompress(d), **msgpack_args)

            s = Sparse(
                data[b"shape"],
                chunks=data[b"chunks"],
                fill_value=data[b"fill_value"],
                dtype=np.dtype(data[b"dtype"]),
                origin=data[b"origin"],
                spacing=data[b"spacing"],
            )

            s._grid = {k: Chunk.from_dict(v) for k, v in data[b"_grid"].items()}

            return s

    def copy(self):
        s = Sparse(
            self.shape,
            chunks=self.chunks,
            fill_value=self.fill_value,
            dtype=self.dtype,
            origin=self.origin,
            spacing=self.spacing,
        )

        s._grid = {
            k: Chunk.from_dict(
                {b"origin": v.origin, b"spacing": v.spacing, b"ndarray": np.copy(v)}
            )
            for k, v in self._grid.items()
        }

        return s

    def astype(self, dtype):
        new_sparse = Sparse.empty_like(self, fill_value=dtype(self.fill_value), dtype=dtype)
        new_sparse.copy_from(self)
        return new_sparse



from multiprocessing.managers import BaseManager
from multiprocessing import Process, JoinableQueue
from time import sleep
import os
import signal
import dill


class ProcessPool:
    """
    Class which enables multiprocess calls to custom functions
    """

    class Shared:
        """
        Object shared between processes. Sync'd by the BaseManager
        """

        def __init__(self):
            self.clear()

        def get(self):
            return self.data

        def add(self, val):
            self.data.append(val)

        def clear(self):
            self.data = []

    def __init__(self, processes_count, *args, **kwargs):
        self.sleep_length = 2
        self.processes_count = processes_count
        self.queue_jobs = JoinableQueue()
        self.processes = []

        BaseManager.register("Shared", self.Shared)
        self.manager = BaseManager()
        self.manager.start()
        self.shared = self.manager.Shared()

        for i in range(self.processes_count):
            p = Process(target=self.make_pool_call)
            p.id = i
            p.start()
            self.processes.append(p)

    def make_pool_call(self):
        while True:
            item_pickled = self.queue_jobs.get()

            if item_pickled is None:
                self.queue_jobs.task_done()
                break

            item = dill.loads(item_pickled)
            call = item.get("call")
            args = item.get("args")
            kwargs = item.get("kwargs")

            try:
                result = call(*args, **kwargs)
                self.shared.add(result)

            except Exception as e:
                import traceback

                traceback.print_exc()
                os.kill(os.getpid(), signal.SIGUSR1)

            self.queue_jobs.task_done()

    def add_job(self, job):
        """
        :param: job: has to be a dilled dict:
                     {
                         'call': function_to_be_called_by_process,
                         'args': [],
                         'kwargs': {},
                     }
        """
        self.queue_jobs.put(job)

    def finish_pool_queue(self):
        while self.queue_jobs.qsize() > 0:
            sleep(self.sleep_length)
        for i in range(self.processes_count):
            self.queue_jobs.put(None)
        self.queue_jobs.join()
        self.queue_jobs.close()
        for p in self.processes:
            p.join()
        del self.processes[:]

    def get_pool_results(self):
        return self.shared.get()

    def clear_pool_results(self):
        self.shared.clear()



from setuptools import setup, find_packages
import os

here = os.path.dirname(os.path.abspath(__file__))


def requirements():
    with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as f:
        return f.read().splitlines()


def version(*parts):
    version_file = os.path.join(*parts)
    version_ns = {}

    with open(version_file) as f:
        exec(f.read(), {}, version_ns)

    return version_ns['__version__']

setup(
    name='chunky3d',
    description='A 3D array-like NumPy-based data structure for large sparsely-populated volumes',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    version=version(here, 'chunky3d', '_version.py'),
    license='MIT',
    install_requires=requirements(),
    extras_require={
        'all': [
            'itk',
            'itk-thickness3d',
            'networkx',
            'scikit-image',
            'SimpleITK',
            'vtk',
        ]
    },
    packages=find_packages(),
    keywords=['3d', 'array', 'chunked', 'sparse'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Multimedia :: Graphics :: 3D Modeling',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ]
    , project_urls={
        'Source': 'https://github.com/K3D-tools/chunky3d',
        'Tracker': 'https://github.com/K3D-tools/chunky3d/issues',
    }
)



import unittest

import numpy as np

from chunky3d.helpers import slice_normalize, slice_shape


class SliceHelper(unittest.TestCase):
    def __getitem__(self, key):
        """Helper magic to easily simulate slicing in [] operator."""
        return key

    def test_hack(self):
        """Testing the behavior of __getitem__."""
        self.assertEqual(self[2, 3], (2, 3))
        self.assertEqual(self[1::3, ...], (slice(1, None, 3), ...))
        self.assertEqual(self[..., :5], (Ellipsis, slice(None, 5)))


class TestSliceNormalize(SliceHelper):
    def test_simple(self):
        self.assertEqual(
            slice_normalize(self[1, 2], (2, 3)), (slice(1, 2, 1), slice(2, 3, 1))
        )

    def test_shape_extension(self):
        self.assertEqual(
            slice_normalize(self[0, 0, ...], (2, 2, 2)),
            (slice(0, 1, 1), slice(0, 1, 1), slice(0, 2, 1)),
        )
        self.assertEqual(
            slice_normalize(self[0, 0], (2, 2, 2)),
            (slice(0, 1, 1), slice(0, 1, 1), slice(0, 2, 1)),
        )
        self.assertEqual(
            slice_normalize(self[0, 0, 0, ...], (2, 2, 2)), (slice(0, 1, 1),) * 3
        )
        self.assertEqual(slice_normalize(self[...], (2, 2, 2)), (slice(0, 2, 1),) * 3)
        self.assertEqual(slice_normalize(self[:], (2, 2, 2)), (slice(0, 2, 1),) * 3)

    def test_overindexing(self):
        with self.assertRaises(IndexError):
            slice_normalize(self[1, 2, 3, 4], (1, 2, 3))

        # ellipsis doesn't overindex
        slice_normalize(self[0, 1, 2, ...], (1, 2, 3))

    def test_ellipse_error(self):
        with self.assertRaises(IndexError):
            slice_normalize(self[..., ...], (1, 2, 3))

    def test_wraparound(self):
        self.assertEqual(slice_normalize(-3, (10,)), (slice(7, 8, 1),))

    def test_out_of_bound(self):
        with self.assertRaises(IndexError):
            slice_normalize(self[3], (3,))

    def test_bad_slice_type(self):
        # slices
        with self.assertRaises(IndexError):
            slice_normalize(self["a":"b"], (1, 2, 3))

        with self.assertRaises(IndexError):
            slice_normalize(self[0:1:"b"], (1, 2, 3))

        # lone indices
        with self.assertRaises(IndexError):
            slice_normalize(self[None], (1, 2, 3))

        with self.assertRaises(IndexError):
            slice_normalize(self["hello"], (1, 2, 3))

        # no exceptions:
        slice_normalize(self[0:None], (1, 2, 3))

    def test_step_prefix(self):
        self.assertEqual(
            slice_normalize(self[::2], (3, 3, 3)),
            (slice(0, 3, 2), slice(0, 3, 1), slice(0, 3, 1)),
        )

    def test_backward_slice(self):
        self.assertEqual(
            slice_normalize(self[::-1], (3, 3, 3)),
            (slice(2, -1, -1), slice(0, 3, 1), slice(0, 3, 1)),
        )

    def test_unsigned_shape(self):
        self.assertEqual(
            slice_normalize(self[1], tuple(np.array([2, 3, 4], dtype=np.uint16))),
            (slice(1, 2, 1), slice(0, 3, 1), slice(0, 4, 1)),
        )


class TestSliceShape(SliceHelper):
    def test_single(self):
        self.assertEqual(slice_shape(self[3], (5,)), (1,))

    def test_empty(self):
        self.assertEqual(slice_shape(self[4:3], (5,)), (0,))

    def test_backward(self):
        self.assertEqual(slice_shape(self[4:3:-1], (5,)), (1,))

    def test_step(self):
        self.assertEqual(slice_shape(self[1:3:2], (5,)), (1,))
        self.assertEqual(slice_shape(self[1:4:2], (5,)), (2,))
        self.assertEqual(slice_shape(self[4:1:-3], (5,)), (1,))

    def test_3d(self):
        self.assertEqual(
            slice_shape(self[::2], (3, 3, 3)),
            (
                2,
                3,
                3,
            ),
        )


if __name__ == "__main__":
    unittest.main()
